from permpy.permutation import Permutation
from permpy.avclass import AvClass
from permpy.permset import PermSet
from permpy.InsertionEncoding.configuration import Configuration
import sympy
import random

class InsertionScheme():
  _tree = {}
  _basis = []
  _max_basis_length = 0
  _automaton = {}
  _root = None
  _configs_checked = set()
  _automaton_ready = False
  _has_inssch = False
  _reductions = {}
  _class = []

  def __init__(self, basis, quiet=False):
    self._basis = basis
    self._max_basis_length = max([len(b) for b in self._basis])
    self._class = AvClass(basis, 1)
    self._configs_checked = set()
    self._automaton_ready = False
    self._root = Configuration((0,), self._basis, 'm')
    self._perm = Configuration((1,), self._basis, 'f')
    self._tree = {self._root: self._root.children()}
    self._automaton = {self._root: list(self._root.children())}
    self._has_inssch = False
    self._reductions = {}
    self._automaton_ready = False
    if self.has_inssch():
      self._has_inssch = True
    else:
      syms = PermSet(self._basis).all_syms()
      for S in syms:
        self._basis = S
        if self.has_inssch():
          self._has_inssch = True
          if not quiet:
            print('Av({}) does not have an insertion scheme, but its symmetry Av({}) does.'.format(basis, S))
          break
      if not self._has_inssch:
        if not quiet:
          print('Neither this class nor its symmetries has an insertion scheme.')


  def has_inssch(self):
    types = [
      PermSet([Permutation(321), Permutation(2143), Permutation(2413)]),
      PermSet([Permutation(123), Permutation(3142), Permutation(3412)]),
      PermSet([Permutation(132), Permutation(312)]),
      PermSet([Permutation(213), Permutation(231)])
    ]
    return all([any([all([b.avoids(p) for p in B]) for b in self._basis]) for B in types])

  def follow_reduce(self,config):
    while config in self._reductions.keys():
      config = self._reductions[config]
    return config


  def build_rules(self, verbose=True, make_class=False, class_bound=100):
    configs_to_check = [Configuration((0,), self._basis)]
    while len(configs_to_check) > 0:
      if verbose:
        # print '\tstates to check:',len(configs_to_check),',   nodes in tree:',len(self._tree.keys()),',   states:', len(self._automaton)
        s_to_print = '\tstates to check: {}, nodes in tree: {}, states: {}'.format(
            len(configs_to_check), len(self._tree.keys()), len(self._automaton))
        print(s_to_print)
      configs_to_check.sort(key=len)
      current_config = configs_to_check[0]
      if verbose:
        print('\nChecking: ',current_config)
      self._configs_checked.add(current_config)

      if current_config.is_permutation():
        if verbose:
          print(current_config,' is not reducible')
        configs_to_check.remove(current_config)
        continue

      length_to_check = self._max_basis_length + current_config.num_slots() - 2

      reducible = False
      order_to_check = sorted(range(len(current_config)), key=lambda i : current_config[i], reverse=True)
      checked_already = list()
      for i in order_to_check:
        if current_config[i] == 0 or (i > 0 and i < len(current_config)-1 and current_config[i-1] == 0 and current_config[i+1] == 0):
          continue
        second_config = Configuration(current_config[:i]+current_config[i+1:], self._basis, '?')
        if second_config in checked_already:
          continue
        checked_already.append(second_config)
        if verbose:
          # print 'Checking isomorphism (depth='+str(length_to_check)+'):',current_config, '<->', second_config
          print('Checking isomorphism (depth={}): {} <-> {}'.format(
            length_to_check, current_config, second_config))
        if self.check_isomorphism(current_config, second_config, length_to_check, make_class, class_bound):
          self._reductions[current_config] = second_config
          if verbose:
            print(current_config,' is reducible to ',second_config)
          if current_config in self._automaton.keys():
            del self._automaton[current_config]
          for config in self._automaton.keys():
            if current_config in self._automaton[config]:
              self._automaton[config].remove(current_config)
              self._automaton[config].append(self.follow_reduce(second_config))
          configs_to_check.append(second_config)
          reducible = True
          break
      if not reducible:
        if verbose:
          print(current_config,'is not reducible')
        if make_class and len(self._class) < len(current_config) + 2 and len(current_config) + 2 <= class_bound:
          if verbose:
            # print '\t\t\tExtending class length from ', (len(self._class) - 1), 'to', (len(current_config) + 2), 'for', current_config, '(!)'
            print('\t\t\tExtending class length from {} to {} for {} (!)'.format(
                len(self._class - 1), len(current_config) + 2, current_config))
          self._class.extend_to_length(len(current_config)+2)
          if verbose:
            print('\t\t\t\tDone!')

        # print '3'
        TTT = current_config.valid_children(self._class)
        self._automaton[current_config] = list(TTT)
        configs_to_check.extend(TTT)

      configs_to_check.remove(current_config)
      configs_to_check = list(set(configs_to_check).difference(self._configs_checked))

    self.standardize_perms()
    self._automaton_ready = True

  def standardize_perms(self):
    for config in self._automaton.keys():
      for c in self._automaton[config]:
        if c.is_permutation():
          self._automaton[config].remove(c)
          self._automaton[config].append(self._perm)


  def check_isomorphism(self, c1, c2, depth, make_class=False, class_bound=100):
    # print '\t\tCI:',depth,c1,c2,len(self._tree.keys())
    if depth == 0 or c1.is_permutation() or c2.is_permutation():
      if c1.is_permutation() != c2.is_permutation():
        pass
        # print c1,c2
      return c1.is_permutation() == c2.is_permutation()
    if c1 not in self._tree.keys():
      if make_class and len(self._class) < len(c1) + 3 and len(c1) + 2 <= class_bound:
        print('\t\t\tExtending class length from ', (len(self._class)-1), 'to', (len(c1) + 2), 'for', c1, '(@)')
        self._class.extend_to_length(len(c1)+2)
        print('\t\t\t\tDone!')
      # print '1',c1,len(c1)
      self._tree[c1] = c1.valid_children(self._class)
    if c2 not in self._tree.keys():
      if make_class and len(self._class) < len(c2) + 3 and len(c2) + 2 <= class_bound:
        print('\t\t\tExtending class length from ', (len(self._class)-1), 'to', (len(c2) + 2), 'for', c2, '(#)')
        self._class.extend_to_length(len(c2)+2)
        print('\t\t\t\tDone!')
      # print '2'
      self._tree[c2] = c2.valid_children(self._class)
    c1kids = {c._type : c for c in self._tree[c1]}
    c2kids = {c._type : c for c in self._tree[c2]}
    if set(c1kids.keys()) != set(c2kids.keys()):
      # print c1,c2,c1kids,c2kids
      return False
    # if random.randint(1,1000) == 535:
      # print '\tnum states:',len(self._tree), '\tdepth:',depth
    return all([self.check_isomorphism(c1kids[i], c2kids[i], depth-1, make_class, class_bound) for i in c1kids.keys()])

  def gf(self, verbose=False, show_series=False):
    if not self._automaton_ready:
      self.build_rules(verbose=verbose)
    states = self._automaton.keys()
    states.append(self._perm)

    x = sympy.Symbol('x')
    transitions = []
    for s1 in states:
      t = []
      for s2 in states:
        if s1 in self._automaton.keys() and s2 in self._automaton[s1]:
          t.append(self._automaton[s1].count(s2)*x)
        else:
          t.append(0)
      transitions.append(t)
    M = (sympy.eye(len(states))-sympy.Matrix(transitions)) ** -1
    f = (1+M[states.index(self._root),states.index(self._perm)]).factor()
    if show_series:
      print(f.series(x,0,11))
    return f




from .permutation import *
from .permset import *
from .permclass import *
from .vectorset import *
from math import factorial,fabs
from sympy import *
from itertools import chain, combinations

def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(0, len(s)+1))

class PegPermutation(Permutation):

  def __new__(cls, p, signs):
    if isinstance(p, int):
      p = list(str(p))
    assert len(p) == len(signs), 'improper sign string'
    for i in range(0, len(signs)):
      assert signs[i] == '+' or signs[i] == '-' or signs[i] == '.', 'improper sign string'

    return Permutation.__new__(cls, p)

  def __init__(self, p, signs):
    self.signs = list(signs)

  def __repr__(self):
    s = ' '
    for i in range(0,len(self)):
      s += str(self[i]+1) + self.signs[i] + ' '
    return s

  def __hash__(self):
    return hash((tuple(self[:]), tuple(self.signs)))

  def __eq__(self, other):
    return hash(self) == hash(other)
    # return super(PegPermutation, self).__eq__(other) and self.signs == other.signs

  def num_signs(self):
    return sum([1 for s in self.signs if s != '.'])

  def sign_subset(self, P):
    for i in range(0, len(self)):
      t = (self.signs[i], P.signs[i])
      if t in [('+', '.'), ('+', '-'), ('-', '+'), ('-', '.')]:
        return False
    return True

  def filling_vector(self):
    L = []
    for s in self.signs:
      if s == '+' or s == '-':
        L.append(2)
      else:
        L.append(1)
    return Vector(L)

  def all_dotted_monotone_intervals(self):
    mi = []
    difference = 0
    c_start = 0
    c_length = 0
    for i in range(0,len(self)-1):
      if self.signs[i] == '.' and self.signs[i+1] == '.' and math.fabs(self[i] - self[i+1]) == 1 and (c_length == 0 or self[i] - self[i+1] == difference):
        if c_length == 0:
          c_start = i
        c_length += 1
        difference = self[i] - self[i+1]
      else:
        if c_length != 0:
          mi.append((c_start, c_start+c_length))
        c_start = 0
        c_length = 0
        difference = 0
    if c_length != 0:
      mi.append((c_start, c_start+c_length))
    return mi

  def is_compact(self):
    for i in range(0, len(self)-1):
      if self[i]+1 == self[i+1]:
        s = ''.join(self.signs[i:i+2])
        if s == '++' or s == '+.' or s == '.+':
          return False
      elif self[i]-1 == self[i+1]:
        s = ''.join(self.signs[i:i+2])
        if s == '--' or s == '-.' or s == '.-':
          return False
    return True

  def is_compact_and_clean(self):
    if not self.is_compact():
      return False
    for i in range(0, len(self)-1):
      if self.signs[i] == '.' and self.signs[i+1] == '.' and fabs(self[i] - self[i+1]) == 1:
        return False
    return True

  def clean_basis(self):
    dotted_intervals = self.all_dotted_monotone_intervals()
    in_intervals = []
    for (start, end) in dotted_intervals:
      in_intervals.extend(range(start, end+1))
    for i in range(0, len(self)):
      if i not in in_intervals:
        dotted_intervals.append((i,i))
    dotted_intervals.sort()
    S = VectorSet()
    for (i,(start, end)) in enumerate(dotted_intervals):
      L = [0]*len(dotted_intervals)
      if start == end:
        if self.signs[start] == '.':
          L[i] = 0
      else:
        length_to_avoid = end - start + 2
        L[i] = length_to_avoid
      for j in L:
        if j != 0:
          if Vector(L).norm() > 0:
            S.append(Vector(L))
          break
    return S

  def split(self, indices):
    entries = list(self[:])
    signs = list(self.signs)
    indices.sort()
    offset = 0
    for index in indices:
      new_entries = entries[:index+offset]
      new_signs = signs[:index+1+offset]
      new_signs.append(signs[index+offset])

      if signs[index+offset] == '+':
        new_entries.append(entries[index+offset]-.5)
        new_entries.append(entries[index+offset]+.5)
      else:
        new_entries.append(entries[index+offset]+.5)
        new_entries.append(entries[index+offset]-.5)        
      new_entries.extend(entries[index+offset+1:])
      new_signs.extend(signs[index+offset+1:])

      entries = Permutation.standardize(new_entries)
      signs = new_signs
      offset += 1
      
    return PegPermutation(entries, signs)

  def clean(self):
    if not self.is_compact():
      return PegPermutation(self, self.signs)
    dotted_intervals = self.all_dotted_monotone_intervals()
    copy = list(self[:])
    copysigns = list(self.signs[:])
    for (start, end) in dotted_intervals:
      type_of_interval = self[start+1]-self[start]
      copysigns[start] = ('+' if type_of_interval == 1 else '-')
      for i in range(start+1, end+1):
        copy[i] = -1
        copysigns[i] = -1
    while -1 in copy:
      copy.remove(-1)
      copysigns.remove(-1)
    return PegPermutation(copy, copysigns)

  def downset(self):
    return PermSet([self]).downset()

  def shrink_by_one(self):
    S = PermSet()

    for i in range(0, len(self)):
      if self.signs[i] != '.':
        new_signs = list(self.signs)
        new_signs[i] = '.'

        S.add(PegPermutation(list(self), new_signs))

      else:
        if len(self) == 1:
          continue
        new_signs = list(self.signs)
        new_signs.pop(i)

        new_entries = list(self)
        new_entries.pop(i)

        S.add(PegPermutation(new_entries, new_signs))

    return S

  def csgf(self, basis):
    x = Symbol('x')
    subsets = powerset(basis)
    f = 0
    for subset in subsets:
      V = VectorSet(subset)
      negpow = pow(-1, len(subset))
      fv = self.filling_vector()
      meet_B = (V.meet_all() if len(V) != 0 else [1]*len(fv))
      xpow = fv.meet(meet_B).norm()
      f += negpow * pow(x,xpow) / pow(1-x,self.num_signs())
    return f

  def reverse(self):
    entries = self[::-1]
    signs = self.signs[::-1]
    for i in range(0,len(signs)):
      if signs[i] == '-':
        signs[i] = '+'
      elif signs[i] == '+':
        signs[i] = '-'
    return PegPermutation(entries, signs)

  def involved_in(self, P):
    if not self.bounds_set:
      (self.lower_bound, self.upper_bound) = self.set_up_bounds()
      self.bounds_set = True
    L = list(self)
    n = len(L)
    p = len(P)
    # if n <= 1 and n <= p:
      # return True

    indices = [0]*n
    
    while indices[0] < p:
      if self.involvement_check(self.upper_bound, self.lower_bound, indices, P, 1):
        newsigns = [P.signs[i] for i in indices]
        if self.sign_subset(P): 
          return True
      indices[0] += 1

    return False

  def involvement_check(self, upper_bound, lower_bound, indices, q, next):
    if next == len(indices):
      return True
    lq = len(q)
    indices[next] = indices[next-1]+1
    while indices[next] < lq:
      if self.involvement_fits(upper_bound, lower_bound, indices, q, next) and self.involvement_check(upper_bound, lower_bound, indices, q, next+1):
        newsigns = [q.signs[i] for i in indices]
        if self.sign_subset(q): 
          return True
      indices[next] += 1
    return False

  def involvement_fits(self, upper_bound, lower_bound, indices, q, next):
    return (lower_bound[next] == -1 or q[indices[next]] > q[indices[lower_bound[next]]]) and (upper_bound[next] == -1 or q[indices[next]] < q[indices[upper_bound[next]]])







from permpy.permutation import *


class Configuration(tuple):
  _basis = set()
  _type = ''
  _children_computed = False
  _children = set()

  def __new__(cls, t, basis=set(), childtype='?'):
    entries = list(t)
    nonzero = [x for x in entries if x != 0]
    assert len(set(nonzero)) == len(nonzero), 'Nonzero entries must be distinct.'
    for i in range(len(entries)-1):
      assert entries[i] != 0 or entries[i+1] != 0, 'Two slots cannot be adjacent.'
    standardization = [0]*len(entries)
    values = sorted(nonzero)
    for i in range(1, len(values)+1):
      standardization[entries.index(values[i-1])] = i
    return tuple.__new__(cls, standardization)

  def __init__(self, t, basis=set(), childtype='?'):
    self._basis = basis
    self._type = childtype
    if childtype == '?':
      t = ''
      if len(self) == 1:
        t = 'f1'
      else:
        loc = self.index(max(self))
        if loc == 0:
          if self[1] == 0:
            t = 'l1'
          else:
            t = 'f1'
        elif loc == len(self) - 1:
          if self[len(self)-2] == 0:
            t = 'r'+str(self.num_slots())
          else:
            t = 'f'+str(self.num_slots()+1)
        else:
          if self[loc-1] == 0 and self[loc+1] == 0:
            t = 'm'+str(len([k for k in self[:loc] if k == 0]))
          elif self[loc-1] == 0:
            t = 'r'+str(len([k for k in self[:loc] if k == 0]))
          elif self[loc+1] == 0:
            t = 'l'+str(len([k for k in self[:loc] if k == 0])+1)
          else:
            t = 'f'+str(len([k for k in self[:loc+1] if k == 0])+1)
      self._type = t

  def num_slots(self):
    return len([x for x in self if x == 0])

  def slot_locs(self):
    return [i for i in range(len(self)) if self[i] == 0]

  def children(self):
    if self._children_computed:
      return self._children
    S = set()
    max_entry = len(self)+1
    L = list(self)
    sl = self.slot_locs()
    for i in sl:
      S.add(Configuration(L[:i]+[max_entry]+L[i+1:], self._basis, 'f'+str(sl.index(i)+1)))
      S.add(Configuration(L[:i+1]+[max_entry]+L[i+1:], self._basis, 'r'+str(sl.index(i)+1)))
      S.add(Configuration(L[:i]+[max_entry]+L[i:], self._basis, 'l'+str(sl.index(i)+1)))
      S.add(Configuration(L[:i+1]+[max_entry]+L[i:], self._basis, 'm'+str(sl.index(i)+1)))
    self._children = S
    self._children_computed = True
    return self._children

  def is_permutation(self):
    return self.num_slots() == 0

  def to_perm(self):
    if self.is_permutation():
      return Permutation(self)
    else:
      return False

  def has_valid_filling(self, C=[]):
    if self.is_permutation():
      return True
    if len(C) - 1 >= self.num_slots():
      filling_entries = C[self.num_slots()]
    else:
      filling_entries = Permutation.listall(self.num_slots())
    for P in filling_entries:
      L = list(self)
      slot_locs = self.slot_locs()
      for (p,v) in enumerate(P):
        L[slot_locs[p]] = v+len(L)
      filled_perm = Permutation(L)
      if len(C) < len(filled_perm) + 1:
        # if len(C) > 2:
          # print 'needed length',(len(filled_perm)),'but only had',(len(C)-1),'[for',self,'/',filled_perm,']'
        if filled_perm.avoids_set(self._basis):
          return True
      else:
        if filled_perm in C[len(filled_perm)]:
          return True
    return False

  def valid_children(self, C=[]):
    return set([c for c in self.children() if c.has_valid_filling(C)])



from __future__ import print_function
from .pegpermutation import *
from .permset import *
from itertools import chain, combinations, combinations_with_replacement
from sympy import *
import time, gc, sys
from sympy.core.cache import clear_cache
import copy


class PegPermSet(PermSet):

  #===== Static Methods to build rearrangement generators =====#

  @staticmethod
  def sortable_by_block_transposition(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 3):
          PP = P.split(list(indices))
          
          entries = PP[:indices[0]+1] + PP[indices[1]+2:indices[2]+3] + PP[indices[0]+1:indices[1]+2] + PP[indices[2]+3:]
          signs = PP.signs[:indices[0]] + PP.signs[indices[1]+1:indices[2]+2] + PP.signs[indices[0]:indices[1]+1] + PP.signs[indices[2]+2:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  @staticmethod
  def sortable_by_prefix_block_transposition(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 2):
          PP = P.split(list(indices))
          
          entries = PP[indices[0]+1:indices[1]+2] + PP[:indices[0]+1] + PP[indices[1]+2:]
          signs = PP.signs[indices[0]+1:indices[1]+2] + PP.signs[:indices[0]+1] + PP.signs[indices[1]+2:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  @staticmethod
  def flip_signs(s):
    for i in range(0,len(s)):
      if s[i] == '-':
        s[i] = '+'
      elif s[i] == '+':
        s[i] = '-'
    return s

  @staticmethod
  def sortable_by_reversal(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 2):
          PP = P.split(list(indices))
          
          entries = PP[:indices[0]+1] + PP[indices[0]+1:indices[1]+2][::-1] + PP[indices[1]+2:]
          signs = PP.signs[:indices[0]+1] + PegPermSet.flip_signs(PP.signs[indices[0]+1:indices[1]+2][::-1]) + PP.signs[indices[1]+2:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  @staticmethod
  def sortable_by_prefix_reversal(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 1):
          PP = P.split(list(indices))
          
          entries = PP[:indices[0]+1][::-1] + PP[indices[0]+1:]
          signs = PegPermSet.flip_signs(PP.signs[:indices[0]+1][::-1]) + PP.signs[indices[0]+1:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  @staticmethod
  def sortable_by_block_interchange(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 4):
          PP = P.split(list(indices))
          
          entries = PP[:indices[0]+1] + \
                      PP[indices[2]+3:indices[3]+4] + \
                      PP[indices[1]+2:indices[2]+3] + \
                      PP[indices[0]+1:indices[1]+2] + \
                      PP[indices[3]+4:]
          signs = PP.signs[:indices[0]+1] + \
                      PP.signs[indices[2]+3:indices[3]+4] + \
                      PP.signs[indices[1]+2:indices[2]+3] + \
                      PP.signs[indices[0]+1:indices[1]+2] + \
                      PP.signs[indices[3]+4:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  @staticmethod
  def sortable_by_cut_and_paste(d):
    assert d == int(d), 'must be given an integer'
    assert d >= 0, 'must be given a nonnegative integer'
    start = PegPermutation(1,'+')
    sortable = PegPermSet([start])
    for i in range(0,d):
      copy = set(sortable)
      for P in copy:
        n = len(P)
        for indices in combinations_with_replacement(range(0, n), 3):
          PP = P.split(list(indices))
          
          entries = PP[:indices[0]+1] + \
                      PP[indices[1]+2:indices[2]+3] + \
                      PP[indices[0]+1:indices[1]+2] + \
                      PP[indices[2]+3:]

          signs = PP.signs[:indices[0]+1] + \
                      PP.signs[indices[1]+2:indices[2]+3] + \
                      PP.signs[indices[0]+1:indices[1]+2] + \
                      PP.signs[indices[2]+3:]

          sortable.add(PegPermutation(entries, signs))

          entries = PP[:indices[0]+1] + \
                      PP[indices[1]+2:indices[2]+3] + \
                      PP[indices[0]+1:indices[1]+2][::-1] + \
                      PP[indices[2]+3:]

          signs = PP.signs[:indices[0]+1] + \
                      PP.signs[indices[1]+2:indices[2]+3] + \
                      PegPermSet.flip_signs(PP.signs[indices[0]+1:indices[1]+2][::-1]) + \
                      PP.signs[indices[2]+3:]

          sortable.add(PegPermutation(entries, signs))

          entries = PP[:indices[0]+1] + \
                      PP[indices[1]+2:indices[2]+3][::-1] + \
                      PP[indices[0]+1:indices[1]+2] + \
                      PP[indices[2]+3:]

          signs = PP.signs[:indices[0]+1] + \
                      PegPermSet.flip_signs(PP.signs[indices[1]+2:indices[2]+3][::-1]) + \
                      PP.signs[indices[0]+1:indices[1]+2] + \
                      PP.signs[indices[2]+3:]

          sortable.add(PegPermutation(entries, signs))
    return sortable

  #===== end Static Methods =====#

  def involves_set(self, P):
    for Q in self:
      if P.involved_in(Q):
        return True
    return False

  def sum_gfs_no_basis(self, S, only_clean=False):
    i = 0
    gf = 0
    n = len(S)
    t = time.time()
    print('\t\tComputing GF.')
    for PP in S:
      i += 1
      if i % 100000 == 0 and i > 0:
        print('\t\t\t',i,'of',n,'\ttook',(time.time()-t),'seconds.')
        t = time.time()
      if not PP.is_compact():
        continue
      if only_clean and not PP.is_compact_and_clean():
        continue
      if i % 10000 == 0 and i > 0:
        gf = gf.simplify()
      if i % 50000 == 0 and i > 0:
        clear_cache()
      gf += PP.csgf([])
      # print 'adding gf for',PP,'with basis []'
      
    return gf

## OLD - Can Probably Be Deleted
  # def alt_downset(self):

  #   topset = PegPermSet(self)

  #   bottom_edge = PegPermSet()
  #   keyssofar = PegPermSet()
  #   unclean = dict()

  #   for PP in self:
  #     if PP.is_compact() and not PP.is_compact_and_clean():
  #       cleaned = PP.clean()
  #       if cleaned in keyssofar:
  #         unclean[cleaned].add(PP)
  #       else:
  #         unclean[cleaned] = PegPermSet([PP])
  #         keyssofar.add(cleaned)

  #   bottom_edge.update(self)
  #   n = len(bottom_edge)
  #   gf = self.sum_gfs_no_basis(bottom_edge, only_clean=True)
  #   while len(bottom_edge) > 0:
  #     oldsize = n
  #     next_layer = bottom_edge.layer_down()
  #     del bottom_edge
  #     clear_cache()
  #     n += len(next_layer)
      
  #     print '\t\tScanning permutations for cleanliness!'
  #     i = 0
  #     num_unclean = 0
  #     nll = len(next_layer)
  #     for PP in next_layer:
  #       i += 1
  #       if i % 200000 == 0:
  #         print '\t\t\tScanned',i,'of',nll,'.'
  #       if PP.is_compact() and not PP.is_compact_and_clean():
  #         cleaned = PP.clean()
  #         num_unclean += 1
  #         if cleaned in keyssofar:
  #           unclean[cleaned].add(PP)
  #         else:
  #           unclean[cleaned] = PegPermSet([PP])
  #           keyssofar.add(cleaned)

  #     print '\t\tScanning permutations for unnecessary uncleans!'
  #     i = 0
  #     nll = len(next_layer)
  #     for PP in next_layer:
  #       i += 1
  #       if i % 200000 == 0:
  #         print '\t\t\tScanned',i,'of',nll,'.'
  #       if unclean.get(PP):
  #         del unclean[PP]

  #     print '\tOut of',len(next_layer),'permutations in this layer,',num_unclean,'were unclean.'
  #     gf += self.sum_gfs_no_basis(next_layer, only_clean=True)

      
  #     bottom_edge = next_layer
  #     del next_layer
  #     clear_cache()
  #     newsize = n
  #     print '\t\tDownset currently has',newsize,'permutations, added',(newsize-oldsize),'in the last run.'

  #   return (gf,unclean)

  def alt_downset(self):

    topset = PegPermSet(self)

    bottom_edge = PegPermSet()
    keyssofar = PegPermSet()
    unclean = dict()

    for PP in self:
      if PP.is_compact() and not PP.is_compact_and_clean():
        cleaned = PP.clean()
        if cleaned in keyssofar:
          unclean[cleaned].add(PP)
        else:
          unclean[cleaned] = PegPermSet([PP])
          keyssofar.add(cleaned)

    bottom_edge.update(self)
    n = len(bottom_edge)
    gf = self.sum_gfs_no_basis(bottom_edge, only_clean=True)

    while len(bottom_edge) > 0:
      oldsize = n
      n = len(bottom_edge)
      next_layer = PegPermSet()

      i = 0
      num_built = 0
      t = time.time()
      while len(bottom_edge) > 0:
        i += 1
        P = bottom_edge.pop()
        next_layer.update(P.shrink_by_one())
        del P
        if i % 100000 == 0:
          clear_cache()
          print('\t',i,'of',n,'. Now with',len(next_layer),'. Took',(time.time()-t),'seconds.')
          t = time.time()


      del bottom_edge
      clear_cache()
      n += len(next_layer)
      
      print('\t\tScanning permutations for cleanliness!')
      i = 0
      num_unclean = 0
      nll = len(next_layer)
      for PP in next_layer:
        i += 1
        if i % 200000 == 0:
          print('\t\t\tScanned',i,'of',nll,'.')
        if PP.is_compact() and not PP.is_compact_and_clean():
          cleaned = PP.clean()
          num_unclean += 1
          if cleaned in keyssofar:
            unclean[cleaned].add(PP)
          else:
            unclean[cleaned] = PegPermSet([PP])
            keyssofar.add(cleaned)

      print('\t\tScanning permutations for unnecessary uncleans!')
      i = 0
      nll = len(next_layer)
      for PP in next_layer:
        i += 1
        if i % 200000 == 0:
          print('\t\t\tScanned',i,'of',nll,'.')
        if unclean.get(PP):
          del unclean[PP]

      print('\tOut of',len(next_layer),'permutations in this layer,',num_unclean,'were unclean.')
      gf += self.sum_gfs_no_basis(next_layer, only_clean=True)

      
      bottom_edge = next_layer
      del next_layer
      clear_cache()
      newsize = n
      print('\t\tDownset currently has',newsize,'permutations.')

    return (gf,unclean)

  def compactify(self):
    copy = PermSet(self)
    for P in copy:
      if not P.is_compact():
        self.remove(P)

  # def downset(self):
  #   return PegPermSet(super(PegPermSet, self).downset())

  def cross_sections(self):
    print('Starting to compute downset.')
    generating_set = PegPermSet(self.downset())
    print('\tDownset done. Contains',len(generating_set),'peg permutations.')
    print('Starting to compactify.')  
    generating_set.compactify()
    print('\tDone compactifying. Now contains',len(generating_set),'peg permutations.')
    print('Starting to compute cross sections.')
    cross_sections = dict()
    pairs = list()
    print('\tCleaning, finding bases, and loading pairs.')
    i = 0
    n = len(generating_set)
    for P in generating_set:
      if i % 20000 == 0:
        print('\t\t',i,'of',n,'...')
      cp = P.clean()
      b = P.clean_basis()
      pairs.append((cp,b))
      cross_sections[cp] = VectorSet([-1])
      i += 1
    i = 0
    del generating_set
    print('\tUnioning bases.')
    for (cleaned_perm, V) in pairs:
      if i % 200000 == 0:
        print('\t\t',i,'of',n,'... dict_size =',len(cross_sections))
      # if cleaned_perm in cross_sections.keys():
      cross_sections[cleaned_perm] = V.basis_union(cross_sections[cleaned_perm])
      # else:
        # cross_sections[cleaned_perm] = V
      i += 1
    del pairs
    return cross_sections

  def enumerate(self, cross_sections=None):
    if cross_sections is None:
      cross_sections = self.cross_sections()
    gc.collect()
    print('\tDone computing cross_sections. There are',len(cross_sections),'cross sections.')
    print('Starting to compute generating function.')
    gf = 0
    i = 0
    n = len(cross_sections)
    t = time.time()
    for clean_perm in cross_sections.keys():
      if i % 10000 == 0 and i > 0:
        gf = gf.simplify()
      if i % 50000 == 0 and i > 0:
        clear_cache()
      if i % 1000 == 0 and i > 0:
        print('\t\t',i,'of',n,'\ttook',(time.time()-t),'seconds.')
        t = time.time()
      gf += clean_perm.csgf(cross_sections[clean_perm])
      i += 1
    print('\tDone!')
    return gf.simplify()

  def alt_cross_sections(self):
    print('Starting to compute downset.')
    (gf, uc) = self.alt_downset()
    unclean = PegPermSet()
    uncleanlist = list()
    for PP in uc.keys():
      uncleanlist.extend(uc[PP])
    unclean = set(uncleanlist )
    print('\tDownset done. Contains',len(unclean),'unclean peg permutations.')
    # print 'Starting to compactify.'
    # unclean.compactify()
    # print '\tDone compactifying. Now contains',len(unclean),'peg permutations.'
    print('Starting to compute cross sections of UNCLEAN peg permutations.')
    cross_sections = dict()
    i = 1
    pairs = list()
    print('\tCleaning, finding bases, and loading pairs.')
    n = len(unclean)
    for P in unclean:
      if i % 20000 == 0:
        print('\t\t',i,'of',n,'...')
      cp = P.clean()
      b = P.clean_basis()
      pairs.append((cp,b))
      cross_sections[cp] = VectorSet([-1])
      i += 1
    i = 1
    del unclean
    clear_cache()
    print('\tUnioning bases.')
    for (cleaned_perm, V) in pairs:
      if i % 200000 == 0:
        print('\t\t',i,'of',n,'... dict_size =',len(cross_sections))
      # if cleaned_perm in cross_sections.keys():
      cross_sections[cleaned_perm] = V.basis_union(cross_sections[cleaned_perm])
      # else:
        # cross_sections[cleaned_perm] = V
      i += 1
    del pairs
    clear_cache()
    return (gf.simplify(), cross_sections)

  def alt_enumerate(self, cross_sections=None):
    '''only works when the set is a generating set for sortables and the top layer has all the same length!!!'''
    ml = max([len(s) for s in self])
    PPS = PegPermSet([s for s in self if len(s) == ml])
    (gf, cross_sections) = PPS.alt_cross_sections()
    gc.collect()
    print('\tDone computing cross_sections. There are',len(cross_sections),'cross sections.')
    print('Starting to compute generating function for uncleans.')
    i = 0
    n = len(cross_sections)
    t = time.time()
    # print 'clean gf:',gf.simplify()
    for clean_perm in cross_sections.keys():
      if i % 10000 == 0 and i > 0:
        gf = gf.simplify()
      if i % 50000 == 0 and i > 0:
        clear_cache()
      if i % 10000 == 0 and i > 0:
        print('\t\t',i,'of',n,'\ttook',(time.time()-t),'seconds.')
        t = time.time()
      # gf -= clean_perm.csgf([])
      # print 'subtracting gf for',clean_perm,'with basis []'
      # print 'adding gf for',clean_perm,'with basis',cross_sections[clean_perm]
      gf += clean_perm.csgf(cross_sections[clean_perm])
      i += 1
    print('\tDone!')
    return gf.simplify()






import random
import fractions
from functools import reduce

import permpy.permutation
from permpy.permutation import Permutation
# import permpy.permclass

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    mpl_imported = True
except ImportError:
    mpl_imported = False


class PermSet(set):
    """Represents a set of permutations, and allows statistics to be computed
    across the set."""

    def __repr__(self):
        return 'Set of {} permutations'.format(len(self))

    def __add__(self, other):
        """Returns the union of the two permutation sets. Does not modify in
        place.

        Example
        -------
        >>> S = PermSet.all(3) + PermSet.all(4)
        >>> len(S)
        30
        """
        result = PermSet()
        result.update(self); result.update(other)
        return result


    @classmethod
    def all(cls, length):
        """Builds the set of all permutations of a given length.

        Parameters:
        -----------
        length : int
            the length of the permutations

        Examples
        --------
        >>> p = Permutation(12); q = Permutation(21)
        >>> PermSet.all(2) == PermSet([p, q])
        True
        """
        return PermSet(Permutation.listall(length))

    def get_random(self):
        """Returns a random element from the set.

        Example
        -------
        >>> p = PermSet.all(4).get_random()
        >>> p in PermSet.all(4) and len(p) == 4
        True
        """

        return random.sample(self, 1)[0]


    def get_length(self, length=None):
        """Returns the subset of permutations which have the specified length.

        Parameters
        ----------
        length : int
            lenght of permutations to be returned

        Example
        -------
        >>> S = PermSet.all(4) + PermSet.all(3)
        >>> S.get_length(3) == PermSet.all(3)
        True
        """
        return PermSet(p for p in self if len(p) == length)

    def heatmap(self, only_length=None, ax=None, blur=False, gray=False, **kwargs):
        """Visalization of a set of permutations, which, for each length, shows
        the relative frequency of each value in each position.

        Paramters
        ---------
        only_length : int or None
            if given, restrict to the permutations of this length
        """
        if not mpl_imported:
            err = 'heatmap requires matplotlib to be imported'
            raise NotImplementedError(err)
        try:
            import numpy as np
        except ImportError as e:
            err = 'heatmap function requires numpy'
            raise e(err)
        # first group permutations by length
        total_size = len(self)
        perms_by_length = {}
        for perm in self:
            n = len(perm)
            if n in perms_by_length:
                perms_by_length[n].add(perm)
            else:
                perms_by_length[n] = PermSet([perm])
        # if given a length, ignore all other lengths
        if only_length:
            perms_by_length = {only_length: perms_by_length[only_length]}
        lengths = list(perms_by_length.keys())
        def lcm(l):
            """Returns the least common multiple of the list l."""
            lcm = reduce(lambda x,y: x*y // fractions.gcd(x,y), l)
            return lcm
        grid_size = lcm(lengths)
        grid = np.zeros((grid_size, grid_size))
        def inflate(a, n):
            """Inflates a k x k array A by into a nk x nk array by inflating
            each entry from A into a n x n matrix."""
            ones = np.ones((n, n))
            c = np.multiply.outer(a, ones)
            c = np.concatenate(np.concatenate(c, axis=1), axis=1)
            return c
        for length, permset in perms_by_length.items():
            small_grid = np.zeros((length, length))
            for p in permset:
                for idx, val in enumerate(p):
                    small_grid[length-val-1, idx] += 1
            mul = grid_size // length
            inflated = inflate(small_grid, mul)
            num_perms = len(permset)
            inflated /= inflated.max()
            grid += inflated

        if not ax:
            ax = plt.gca()
        if blur:
            interpolation = 'bicubic'
        else:
            interpolation = 'nearest'
        def get_cubehelix(gamma=1, start=1, rot=-1, hue=1, light=1, dark=0):
            """Get a cubehelix palette."""
            cdict = mpl._cm.cubehelix(gamma, start, rot, hue)
            cmap = mpl.colors.LinearSegmentedColormap("cubehelix", cdict)
            x = np.linspace(light, dark, 256)
            pal = cmap(x)
            cmap = mpl.colors.ListedColormap(pal)
            return cmap
        if gray:
            cmap = get_cubehelix(start=.5, rot=1, light=1, dark=.2, hue=0)
        else:
            cmap = get_cubehelix(start=.5, rot=-.5, light=1, dark=.2)

        ax.imshow(grid, cmap=cmap, interpolation=interpolation)
        ax.set_aspect('equal')
        ax.set(**kwargs)
        ax.axis('off')
        return ax











    def show_all(self):
        """The default representation doesn't print the entire set, this
        function does."""
        return set.__repr__(self)

    def __add__(self, other):
        result = PermSet()
        result.update(self)
        result.update(other)
        return result

    def minimal_elements(self):
        """Returns the elements of the set which are minimal with respect to
        the permutation pattern order.

        Examples
        --------
        """



        B = list(self)
        B = sorted(B, key=len)
        C = B[:]
        n = len(B)
        for (i,b) in enumerate(B):
            # if i % 1 == 0:
                # print i,'/',n
            if b not in C:
                continue
            for j in range(i+1,n):
                if B[j] not in C:
                    continue
                if b.involved_in(B[j]):
                    C.remove(B[j])
        return PermSet(C)

    def all_syms(self):
        sym_set = [frozenset(self)]
        sym_set.append(frozenset([i.reverse() for i in self]))
        sym_set.append(frozenset([i.complement() for i in self]))
        sym_set.append(frozenset([i.reverse().complement() for i in self]))
        sym_set.extend([frozenset([k.inverse() for k in L]) for L in sym_set])
        return frozenset(sym_set)

    def layer_down(self):
        S = PermSet()
        i = 1
        n = len(self)
        for P in self:
            # if i % 10000 == 0:
                # print('\t',i,'of',n,'. Now with',len(S),'.')
            S.update(P.shrink_by_one())
            i += 1
        return S

    def downset(self, return_class=False):
        bottom_edge = PermSet()
        bottom_edge.update(self)

        done = PermSet(bottom_edge)
        while len(bottom_edge) > 0:
            oldsize = len(done)
            next_layer = bottom_edge.layer_down()
            done.update(next_layer)
            del bottom_edge
            bottom_edge = next_layer
            del next_layer
            newsize = len(done)
            # print '\t\tDownset currently has',newsize,'permutations, added',(newsize-oldsize),'in the last run.'
        if not return_class:
            return done
        cl = [PermSet([])]
        max_length = max([len(P) for P in done])
        for i in range(1,max_length+1):
            cl.append(PermSet([P for P in done if len(P) == i]))
        return permclass.PermClass(cl)


    def total_statistic(self, statistic):
        return sum([statistic(p) for p in self])

    def threepats(self):
        patnums = {'123' : 0, '132' : 0, '213' : 0,
                             '231' : 0, '312' : 0, '321' : 0}
        L = list(self)
        for p in L:
            n = len(p)
            for i in range(n-2):
                for j in range(i+1,n-1):
                    for k in range(j+1,n):
                        std = permpy.permutation.Permutation.standardize([p[i], p[j], p[k]])
                        patnums[''.join([str(x + 1) for x in std])] += 1
        return patnums

    def fourpats(self):
        patnums = {'1234' : 0, '1243' : 0, '1324' : 0,
                             '1342' : 0, '1423' : 0, '1432' : 0,
                             '2134' : 0, '2143' : 0, '2314' : 0,
                             '2341' : 0, '2413' : 0, '2431' : 0,
                             '3124' : 0, '3142' : 0, '3214' : 0,
                             '3241' : 0, '3412' : 0, '3421' : 0,
                             '4123' : 0, '4132' : 0, '4213' : 0,
                             '4231' : 0, '4312' : 0, '4321' : 0 }
        L = list(self)
        for p in L:
            n = len(p)
            for i in range(n-3):
                for j in range(i+1,n-2):
                    for k in range(j+1,n-1):
                        for m in range(k+1,n):
                            std = permpy.permutation.Permutation.standardize([p[i], p[j], p[k], p[m]])
                            patnums[''.join([str(x + 1) for x in std])] += 1
        return patnums



import copy
import time
from math import factorial

import permpy.permutation
from permpy.permset import PermSet

class PermClass(list):

    # def __init__(self, n = 8):
        # list.__init__(self, [permset.PermSet(permutation.Permutation.listall(i)) for i in range(n + 1)])
        # self.avoids = []
        # self.length = n

    # def __len__(self):
    #       return self.length

    @staticmethod
    def class_from_test(test, l=8, has_all_syms=False):
        C = [permset.PermSet([permutation.Permutation([])])]
        for cur_length in range(1,l+1):
            this_len = permset.PermSet([])
            if len(C[cur_length-1]) == 0:
                return PermClass(C)
            to_check = permset.PermSet(set.union(*[P.all_extensions() for P in C[cur_length-1]]))
            to_check = [P for P in to_check if permset.PermSet(P.children()).issubset(C[cur_length-1])]
            while len(to_check) > 0:
                P = to_check.pop()
                print(str(P))
                if has_all_syms:
                    syms = permset.PermSet([
                            P,
                            P.reverse(),
                            P.complement(),
                            P.reverse().complement(),
                            P.inverse(),
                            P.reverse().inverse(),
                            P.complement().inverse(),
                            P.reverse().complement().inverse()
                        ])
                if test(P):
                    if has_all_syms:
                        for Q in syms:
                            this_len.add(Q)
                    else:
                        this_len.add(P)
                if has_all_syms:
                    for Q in syms:
                        if Q in to_check:
                            to_check.remove(Q)

            C.append(this_len)
        return PermClass(C)



    def filter_by(self, test):
        for i in range(0, len(self)):
            D = list(self[i])
            for P in D:
                if not test(P):
                    self[i].remove(P)

    def guess_basis(self, max_length=6, search_mode=False):
        """
            Guess a basis for the class up to "max_length" by iteratively generating
            the class with basis elements known so far ({}, to start with) and adding
            elements which should be avoided to the basis.

            Search mode goes up to the max length in the class and prints out the number
            of basis elements of each length on the way.
        """

        t = time.time()

        assert max_length < len(self), 'class not big enough to check that far'

        if search_mode:
            max_length = len(self)-1

        # Find the first length at which perms are missing.
        not_all_perms = [i for i in range(len(self)) if i >= 1 and len(self[i]) != factorial(i)]

        # If no perms are missing, we have all perms, so return empty basis.
        if len(not_all_perms) == 0:
            return permset.PermSet([])

        # Add missing perms of minimum length to basis.
        start_length = min(not_all_perms)
        basis = permset.PermSet(permutation.Permutation.listall(start_length)).difference(self[start_length])

        if search_mode:
            print('\t'+str(len(basis))+' basis elements of length '+str(start_length)+'\t\t'+("{0:.2f}".format(time.time() - t)) + ' seconds')
            t = time.time()

        basis_elements_so_far = len(basis)

        current_length = start_length + 1

        # Go up in length, adding missing perms at each step.
        while current_length <= max_length:
            C = avclass.AvClass(basis, current_length)
            basis = basis.union(C[-1].difference(self[current_length]))

            if search_mode:
                print('\t'+str(len(basis)-basis_elements_so_far)+' basis elements of length ' + str(current_length) + '\t\t' + ("{0:.2f}".format(time.time() - t)) + ' seconds')
                t = time.time()

            basis_elements_so_far = len(basis)

            current_length += 1

        return basis


    # def guess_basis(self, max_length=8):
    #       max_length = min(max_length, len(self)-1)
    #       B = permset.PermSet()
    #       B.update(self.check_tree_basis(max_length, permutation.Permutation([1,2]), permset.PermSet()))
    #       B.update(self.check_tree_basis(max_length, permutation.Permutation([2,1]), permset.PermSet()))
    #       return B.minimal_elements()

    # def check_tree_basis(self, max_length, R, S):
    #       if R not in self[len(R)]:
    #           for s in S:
    #               if s.involved_in(R):
    #                   return S
    #           S.add(R)
    #           return S
    #       else:
    #           if len(R) == max_length:
    #               return S
    #           re = R.right_extensions()
    #           for p in re:
    #               S = self.check_tree_basis(max_length, p, S)
    #           return S

    def plus_class(self,t):
        C = copy.deepcopy(self)
        for i in range(0,t):
            C = C.plus_one_class()
        return C

    def plus_one_class(self):
        D = copy.deepcopy(self)
        D.append(permset.PermSet())
        for l in range(0,len(self)):
            for P in self[l]:
                D[l+1] = D[l+1].union(P.all_extensions())
        return D

    def heatmap(self, **kwargs):
        permset = PermSet()
        for item in self:
            permset.update(item)
        permset.heatmap(**kwargs)

    def sum_closure(self,length=8, has_syms=False):
        return PermClass.class_from_test(lambda P : ((len(P) < len(self) and P in self[len(P)]) or P.sum_decomposable()) and all([Q in self[len(Q)] for Q in P.chom_sum()]), l=length, has_all_syms=has_syms)





from .vector import *

class VectorSet(list):

  def basis_union(self, B):
    if len(B) > 0 and B[0] == -1:
      return self
    if len(self) > 0 and self[0] == -1:
      return B
    S = set()
    for V in self:
      for W in B:
        S.add(V.meet(W))
    return VectorSet(S).minimal_elements()

  def minimal_elements(self):
    C = self[:]
    for V in self:
      for W in self:
  	    if V == W:
  	      continue
  	    if W.contained_in(V):
  	      C.remove(V)
  	      break  
    return C

  def meet_all(self):
    if len(self) == 0:
      return Vector([])

    l = len(self[0])
    V = Vector([1]*l)
    for W in self:
      V = V.meet(W)
    return V



import itertools as it


# Generic Permutation class, others build from this?



class _GeneralPermutation(tuple):
  
  # determines whether permutations are indexed from 0 or 1 
  _startIndex = 1

  def __new__(cls, p, distinct = True):
    ''' Initializes a mpermutation object, internal indexing starts at zero. '''
    if isinstance(p, cls):
      return tuple.__new__(cls, p)
    elif isinstance(p, tuple):
      entries = list(p)[:]
    elif isinstance(p, list):
      entries = list(set(p))
    # standardizes, starting at zero
    if distinct:
      assert len(entries) == len(p)
    entries.sort()
    standardization =  map(lambda e: entries.index(e), p)
    return tuple.__new__(cls, standardization)

  def oneline_repr(self):
    s = ' '
    for i in self:
      s += str(i +  self._startIndex) + ' '
    return s

  def __repr__(self):
    return self.oneline_repr()

  def __call__(self,i):
    '''allows permutations to be used as functions 
    (useful for counting cycles)'''
    return self[i]

  def delete(self, idx):
    L = list(self)
    del L[idx]
    return self.__class__(L)




class MPerm(genperm._GeneralPermutation):
  ''' Class representing multiset permutations '''
  
  # static class variable, controls permutation representation
  _REPR = 'oneline'  

  def __new__(cls, p):
    return super(MPerm, cls).__new__(cls, p, distinct = False)








# import permpy
from .permset import PermSet
from .permutation import Permutation
from .pegpermutation import PegPermutation
from .permclass import PermClass
from .avclass import AvClass
from .pegpermset import PegPermSet
from .geometricgridclass import GeometricGridClass
from .InsertionEncoding import *
import permpy.RestrictedContainer

import doctest
doctest.testmod()




from .permset import PermSet
from .permutation import Permutation
from .pegpermutation import PegPermutation
from .permclass import PermClass
from .avclass import AvClass
from .pegpermset import PegPermSet
from .geometricgridclass import GeometricGridClass
from .InsertionEncoding import *
import permpy.RestrictedContainer

try:
    import matplotlib as mpl
    mpl.rc('axes', fc='E5E5E5', ec='white', lw='1',
            grid='True', axisbelow='True')
    mpl.rc('grid', c='white', ls='-')
    mpl.rc('figure', fc='white')
    mpl_imported = True
except ImportError:
    print('Install matplotlib for extra plotting functionality')
    pass


Perm = Permutation
Av = AvClass



''' This is working code which may or may not be added to a class at some point.'''

from __future__ import print_function
from permpy import *
from math import factorial
import itertools
from collections import Counter
from permpy.RestrictedContainer import *


def greedy_sum(p):
  parts = []
  sofar = 0
  while sofar < len(p):
    if len(p)-sofar == 1:
      parts.append(Perm(1))
      return parts
    i = 1
    while sofar+i <= len(p) and list(p[sofar:sofar+i]) == range(sofar,sofar+i):
      i += 1
    i -= 1
    if i > 0:
      parts.append(Perm(range(i)))
    sofar += i
    i = 2
    while sofar+i <= len(p) and not (max(p[sofar:sofar+i]) - min(p[sofar:sofar+i])+1 == i and min(p[sofar:sofar+i]) == sofar):
      i += 1
    if sofar+i <= len(p):
      parts.append(Perm(p[sofar:sofar+i]))
    sofar += i
  return parts

def chom_sum(p):
  L = []
  p = greedy_sum(p)
  for i in p:
    if i.inversions() == 0:
      L.extend([Perm(1)]*len(i))
    else:
      L.append(i)
  return L

def chom_skew(p):
  return [r.reverse() for r in chom_sum(p.reverse())]

def expected_basis(B):
  return PermSet([expected_basis_element(P) for P in B]).minimal_elements()

def expected_basis_element(P):
  if P[0] == len(P)-1:
    return P
  else:
    return Permutation([len(P)+2]+list(P))

def powerset(iterable):
  s = list(iterable)
  return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s)+1))

def check_combos(max_length, size_of_combos=0, verbose=False):
  bad_ones = 0
  so_far = 0
  permlist = []
  for i in range(2,max_length+1):
    permlist.extend(PermSet.all(i))
  if size_of_combos == 0:
    to_check = powerset(permlist)
  else:
    to_check = itertools.combinations(permlist, size_of_combos)
  for basis in to_check:
    if len(basis) < 1:
      continue
    so_far += 1
    print('----------------------------------------------------')
    expected = expected_basis(basis)
    B = make_restricted_container_class(basis).guess_basis()
    if B == expected:
      print(str(basis) + '  -->  ' + str(B) + '  :  ok')
    else:
      bad_ones += 1
      print(str(basis) + '  -->  ' + str(B) + '  : WHOA!')
    if verbose and so_far % 10 == 0:
      print('----------------------------------------------------')
      print('----------------------------------------------------')
      print('       === SO FAR: ' + str(bad_ones) + ' BAD BASES OUT OF ' + str(so_far) + ' ===')
      print('----------------------------------------------------')
  print('----------------------------------------------------')
  print('----------------------------------------------------')
  print('----------------------------------------------------')
  print('=== THERE WERE ' + str(bad_ones) + ' BAD BASES OUT OF ' + str(so_far) + ' ===')
  print('----------------------------------------------------')


def check_singleton_restricted_containers():
  for length in range(2,6):
    print('----------------------------------------------------')
    for index in range(0, factorial(length)):
      print('----------------------------------------------------')
      P = Permutation(index, length)
      expected = expected_basis(P)
      B = make_restricted_container_class([P]).guess_basis()
      if B == expected:
        print(str(P) + '  -->  ' + str(B) + '  :  ok')
      else:
        print(str(P) + '  -->  ' + str(B) + '  : WHOA!')
  print('----------------------------------------------------')

def make_restricted_container_class(basis,length=7):
  basis = [Permutation(b) for b in basis]
  D = [PermSet([])]
  for i in range(1, length+1):
    A = RestrictedContainer(basis,i)
    D.append(generate_from_restricted_container(A))
  return PermClass(D)

def generate_from_restricted_container(container):
  perms_generated = PermSet()
  queue = [container]
  while len(queue) > 0:
    machine = queue.pop(0)
    container_size = machine.container_size()
    input_size = machine.input_size()
    if container_size == 0 and input_size == 0:
      output = machine.output()
      output_permutation = Permutation(output)
      perms_generated.add(output_permutation)
    else:
      if container_size > 0:
        queue.append(machine.pop())
      for open_spot in range(0, container_size+1):
        new_machine = machine.push(open_spot)
        if new_machine != -1:
          queue.append(new_machine)
  return perms_generated


def color_4231(p):
  red_entries = []
  blue_entries = []

  for entry in p:
    if Permutation(blue_entries+[entry]).involves(Permutation(312)):
      red_entries.append(entry)
    else:
      blue_entries.append(entry)

  return (red_entries, blue_entries)

def label_4231(p, new_version=False):
  (red_entries, blue_entries) = color_4231(p)
  word = ''
  for (index, entry) in enumerate(p):
    if new_version and index in p.ltrmax():
      word += 'D'
    elif new_version and index in p.rtlmin():
      word += 'A'
    elif entry in red_entries:
      word += ('A' if index in p.rtlmin() else 'B')
    else:
      word += ('D' if index in p.ltrmax() else 'C')
  return (word, ''.join([word[i] for i in p.inverse()]))

def color_1324(p):
  red_entries = []
  blue_entries = []

  for entry in p:
    if (len(blue_entries) > 0 and entry > min(blue_entries)) or Permutation(red_entries+[entry]).involves(Permutation(132)):
      blue_entries.append(entry)
    else:
      red_entries.append(entry)

  return (red_entries, blue_entries)

def label_1324(p, new_version=False):
  (red_entries, blue_entries) = color_1324(p)
  word = ''
  for (index, entry) in enumerate(p):
    if new_version and index in p.rtlmax():
      word += 'D'
    elif entry in red_entries:
      word += ('A' if index in p.ltrmin() else 'B')
    else:
      word += ('D' if index in p.rtlmax() else 'C')
  return (word, ''.join([word[i] for i in p.inverse()]))

def check_pattern(word_pairs, pattern):
  n = [0,0]
  for (x,y) in word_pairs:
    if x.find(pattern) != -1:
      n[0] += 1
    if y.find(pattern) != -1:
      n[1] += 1
  if n[0] == 0 or n[1] == 0:
    print('\n\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n\t\t'+pattern+': ('+str(n[0])+', '+str(n[1])+')\n\n')
  return n

def check_pattern_list(word_pairs, patterns):
  d = dict()
  for pattern in patterns:
    d[pattern] = check_pattern(word_pairs, pattern)
  return d

def allstrings(alphabet, length):
  c = []
  for i in range(length):
    c = [[x]+y for x in alphabet for y in c or [[]]]
  return [''.join(a) for a in c]

def nc_contains(w,u):
  seen = 0
  for i in range(0, len(w)):
    if w[i] == u[seen]:
      seen += 1
      if seen == len(u):
        return True
  return False

def check_nc_pattern(word_pairs, pattern):
  n = [0,0]
  for (x,y) in word_pairs:
    if nc_contains(x,pattern):
      n[0] += 1
    if nc_contains(y,pattern):
      n[1] += 1
  return n

def check_nc_pattern_list(word_pairs, patterns):
  d = dict()
  for pattern in patterns:
    d[pattern] = check_nc_pattern(word_pairs, pattern)
  return d


# implements rsk:
class Tab(object):
  def __init__(self,P,Q):
    self.P = P
    self.Q = Q

  def __repr__(self):
    l = len(self.P[0])
    h = len(self.P)
    sP = 'P = |'
    sQ = 'Q = |'
    for i in range(h):
      for j in range(len(self.P[i])):
        sP += ('%2i' %   self.P[i][j]) + '|'
        sQ += ('%2i' %   self.Q[i][j]) + '|'
      sP +=   '\n    |'
      sQ +=   '\n    |'
    return sP[:-6] +'\n'+ sQ[:-6]

def rsk(perm):
  p = [i + 1 for i in perm[::-1]]
  q = 1
  P = [[]]
  Q = [[]]
  def insert(k,r,q1):
    if len(P) < r+1:
      P.append([])
      Q.append([])
    if not P[r] or k > P[r][-1]:
      P[r].append(k)
      Q[r].append(q1)
    else:
      n = len(P[r])
      i = n-1
      while True:
        b = P[r][i]
        if k < b:
          if i >0:
            i -=1
          else:
            P[r][0] = k
            break
        if k > b:
          b = P[r][i+1]
          P[r][i+1] = k
          break
      insert(b,r+1,q)
  for i in range(len(perm)-1,-1,-1):
    insert(p.pop(),0,q)
    q+=1
  return Tab(P,Q)
  
def rsk_shape(perm):
	P = rsk(perm).P
	return([len(P[i]) for i in range(0,len(P))])
	
def shape_contains(A,B):
	return(len(A) >= len(B) and all(B[i] <= A[i] for i in range(0,len(B))))

def check_stat(f, L1, L2, l=8):
  C = [sorted([f(P) for P in T[l]]) for T in L1]
  D = [sorted([f(P) for P in T[l]]) for T in L2]
  return [(i,j) for i in range(len(L1)) for j in range(len(L2)) if C[i] == D[j]]

def check_stat_list(fs, L1, L2, l=8):
  C = [sorted([tuple([f(P) for f in fs]) for P in T[l]]) for T in L1]
  D = [sorted([tuple([f(P) for f in fs]) for P in T[l]]) for T in L2]
  # done = []
  # vals = []
  # for thing in C[0]:
    # if thing in done:
      # continue
    # done.append(thing)
    # vals.append(C[0].count(thing))
    # print C[0].count(thing),'\t->\t',thing
  # print max(vals)
  return [(i,j) for i in range(len(L1)) for j in range(len(L2)) if C[i] == D[j]]

def size_value_blocks(L):
  if len(L) == 0:
    return []
  L = sorted(list(L))
  T = []
  s = 0
  for i in range(len(L)-1):
    s += 1
    if L[i]+1 not in L:
      T.append(s)
      s = 0
  s += 1
  T.append(s)
  return sorted(T)

def rhs_blocks(P):
  right_side = P[P.index(0):]
  rhs_sorted = sorted(right_side)
  sorted_lookup = sorted(enumerate(right_side), key=lambda i:i[1])
  rhs_blocks = []
  b_sizes = []
  this_block_min = 0
  this_block_size = 1
  for i in range(len(rhs_sorted)-1):
    if rhs_sorted[i] + 1 in rhs_sorted:
      this_block_size += 1
    else:
      rhs_blocks.append([t[1] for t in sorted([sorted_lookup[i] for i in range(this_block_min,this_block_min+this_block_size)],key=lambda i:i[0])])
      this_block_min = i+1
      b_sizes.append(this_block_size)
      this_block_size = 1
  b_sizes.append(this_block_size)
  # if this_block_size == 1:

  rhs_blocks.append([t[1] for t in sorted([sorted_lookup[i] for i in range(this_block_min,this_block_min+this_block_size)],key=lambda i:i[0])])
  return rhs_blocks

def size_LR_max_contig_blocks(P):
  rhs_bs = rhs_blocks(P)
  sizes = []
  for i in range(1, len(rhs_bs)):
    b = rhs_bs[i]
    P = Perm(b)
    vs = [P[i] for i in P.ltrmax()]
    v = max(vs)
    j = 1
    while v-j in vs:
      j += 1
    sizes.append(j)
  return sizes


def num_value_blocks(L):
  if len(L) == 0:
    return 0
  L = sorted(list(L))
  n = 1
  for i in range(len(L)-1):
    if L[i]+1 not in L:
      n += 1
  return n

def longest_init_asc(L):
  n = 1
  for i in range(1, len(L)):
    if L[i] > L[i-1]:
      n += 1
    else:
      return n
  return n

def longest_init_desc(L):
  return longest_init_asc(Perm(L).complement())

def split_and_check_perm(P):
  (a,b) = (P[:P.index(0)], P[P.index(0):])
  return (num_value_blocks(a), num_value_blocks(b))

def split_size_value_blocks(P):
  (a,b) = (P[:P.index(0)], P[P.index(0):])
  return (size_value_blocks(a), size_value_blocks(b))

def up_jump_lengths(P):
	L = []
	for i in range(len(P)-1):
		if P[i+1]-P[i] > 0:
			L.append(P[i+1]-P[i])
		else:
			L.append(0)
	return L


def check_stats(L1, L2, l=8, add_syms=False, tups=1):
  if add_syms:
    L1 = [item for sublist in [list(PermSet(l1).all_syms()) for l1 in L1] for item in sublist]
    L2 = [item for sublist in [list(PermSet(l2).all_syms()) for l2 in L2] for item in sublist]
  stats = [
    ('number of descents', Perm.descents),
    ('position of descents', lambda P : [i for i in range(len(P)-1) if P[i] > P[i+1]]),
    ('number of ascents', Perm.ascents),
    ('position of ascents', lambda P : [i for i in range(len(P)-1) if P[i] < P[i+1]]),
    ('number of ltrmin', lambda P : len(P.ltrmin())),
    ('positions of ltrmin', Perm.ltrmin),
    ('values of ltrmin', lambda P : [P[i] for i in P.ltrmin()]),
    ('number of ltrmax', lambda P : len(P.ltrmax())),
    ('positions of ltrmax', Perm.ltrmax),
    ('values of ltrmax', lambda P : [P[i] for i in P.ltrmax()]),
    ('number of rtlmin', lambda P : len(P.rtlmin())),
    ('positions of rtlmin', Perm.rtlmin),
    ('values of rtlmin', lambda P : [P[i] for i in P.rtlmin()]),
    ('number of rtlmax', lambda P : len(P.rtlmax())),
    ('positions of rtlmax', Perm.rtlmax),
    ('values of rtlmax', lambda P : [P[i] for i in P.rtlmax()]),
    ('number of sum decomposables', Perm.sum_decomposable),
    ('number of sum components', lambda P: len(chom_sum(P))),
    ('number of skew decomposables', Perm.skew_decomposable),
    ('number of skew components', lambda P: len(chom_skew(P))),
    ('number of inflations of simples', lambda P : P.skew_decomposable() or P.sum_decomposable()),
    ('length of simple quotient', lambda P : len(P.decomposition()[0])),
    ('number of simple permutations', Perm.is_simple),
    ('number of bonds', Perm.bonds),
    ('+ bonds', lambda P: len([i for i in range(len(P)-1) if P[i+1] == P[i]+1])),
    ('- bonds', lambda P: len([i for i in range(len(P)-1) if P[i+1] == P[i]-1])),
    ('position of + bonds', lambda P : [i for i in range(len(P)-1) if P[i]+1==P[i+1]]),
    ('position of - bonds', lambda P : [i for i in range(len(P)-1) if P[i]-1==P[i+1]]),
    ('number of bends', Perm.bends),
    ('position of bends', lambda P : [i for i in range(1, len(P)-1) if (P[i-1] > P[i] and P[i+1] > P[i]) or (P[i-1] < P[i] and P[i+1] < P[i])]),
    ('number of valleys', lambda P : len([i for i in range(1, len(P)-1) if P[i-1] > P[i] and P[i+1] > P[i]])),
    ('position of valleys', lambda P : [i for i in range(1, len(P)-1) if P[i-1] > P[i] and P[i+1] > P[i]]),
    ('number of peaks', lambda P : len([i for i in range(1, len(P)-1) if P[i-1] < P[i] and P[i+1] < P[i]])),
    ('position of peaks', lambda P : [i for i in range(1, len(P)-1) if P[i-1] < P[i] and P[i+1] < P[i]]),
    ('number of exceedances', lambda P : len([i for i in range(len(P)) if P[i] > i])),
    ('position of exceedances', lambda P : [i for i in range(len(P)) if P[i] > i]),
    ('value of first entry', lambda P : P[0]),
    ('value of second entry', lambda P : P[1]),
    ('value of second to last entry', lambda P : P[len(P)-2]),
    ('value of last entry', lambda P : P[len(P)-1]),
    ('position of 1', lambda P : [i for i in range(len(P)) if P[i] == 0]),
    ('position of 2', lambda P : [i for i in range(len(P)) if P[i] == 1]),
    ('position of n-1', lambda P : [i for i in range(len(P)) if P[i] == len(P)-2]),
    ('position of n', lambda P : [i for i in range(len(P)) if P[i] == len(P)-1]),
    ('major index', Perm.majorindex),
    ('longest ascending run', Perm.longestrunA),
    ('longest descending run', Perm.longestrunD),
    ('longest run', Perm.longestrun),
    ('number of inversions', Perm.inversions),
    ('number of top steps', lambda P : len([i for i in [i for i in range(len(P)-1) if P[i] > P[i+1]] if i in P.rtlmax() and i+1 in P.rtlmax() and P[i]-1==P[i+1]])),
    ('position of top steps', lambda P : [i for i in [i for i in range(len(P)-1) if P[i] > P[i+1]] if i in P.rtlmax() and i+1 in P.rtlmax() and P[i]-1==P[i+1]]),
    ('number of fixed points', lambda P : len([i for i in range(len(P)) if P[i] == i])),
    ('position of fixed points', lambda P : [i for i in range(len(P)) if P[i] == i]),
    ('number of disjoint cycles', Perm.num_disjoint_cycles),
    ('number of involutions', Perm.is_involution),
    ('rank encodings', Perm.rank_encoding),
    ('max rank', lambda P : max(P.rank_encoding())),
    # ('num rtlmax / ltrmin layers', Perm.num_rtlmax_ltrmin_layers),
    # ('num ltrmax / rtlmin layers', lambda P : P.reverse().num_rtlmax_ltrmin_layers()),
    # ('num rtlmax / ltrmin layers contracted', lambda P : P.contract_inc_bonds().num_rtlmax_ltrmin_layers()),
    ('occurrences of 123', lambda P : P.occurrences(Perm(123))),
    ('occurrences of 132', lambda P : P.occurrences(Perm(132))),
    ('occurrences of 213', lambda P : P.occurrences(Perm(213))),
    ('occurrences of 231', lambda P : P.occurrences(Perm(231))),
    ('occurrences of 312', lambda P : P.occurrences(Perm(312))),
    ('occurrences of 321', lambda P : P.occurrences(Perm(321))),
    ('occurrences of 2413', lambda P : P.occurrences(Perm(2413))),
    # ('split and count value-blocks', split_and_check_perm),
    # ('size_LR_max_contig_blocks', size_LR_max_contig_blocks),
  ]

  C1 = [Av(B,l) for B in L1]
  C2 = [Av(B,l) for B in L2]

  print("")
  if tups == 1:
    for (name, f) in stats:
      equivs = check_stat(f, C1, C2, l)
      print(name,":")
      if len(equivs) == 0:
        print("\tnone\n")
        continue
      for (x, y) in equivs:
        print("\t",L1[x],"~",L2[y])
      print("")
  else:
    fs = itertools.combinations(stats, tups)
    for stat_set in fs:
      equivs = check_stat_list([stat_set[i][1] for i in range(len(stat_set))], C1, C2, l)
      print("")
      print('[',(' / '.join([stat_set[i][0] for i in range(len(stat_set))])),']')
      if len(equivs) == 0:
        print("\tnone\n")
        continue
      for (x, y) in equivs:
        print("\t",L1[x],"~",L2[y])

def kill_syms(bases):
  if len(bases) == 0:
    return bases
  bases = list(bases)
  new_bases = [PermSet(bases[0])]

  for B in bases[1:]:
    syms = [PermSet(K) for K in PermSet(B).all_syms()]
    if not any([BSym in new_bases for BSym in syms]):
      new_bases.append(PermSet(B))
  return new_bases

def kill_auts(basis_so_far, additions):
  auts = [
    lambda P : P.reverse(),
    lambda P : P.complement(),
    lambda P : P.reverse().complement(),
    lambda P : P.inverse(),
    lambda P : P.inverse().reverse(),
    lambda P : P.inverse().complement(),
    lambda P : P.inverse().reverse().complement()
  ]

  good_auts = [a for a in auts if all([P == a(P) for P in basis_so_far])]

  keep_additions = []
  for beta in additions:
    if not any([a(beta) in keep_additions for a in good_auts]):
      keep_additions.append(beta)

  return keep_additions
  


def hunt_for_enumeration(enum, max_len = 8, check_ahead = 2, verbose = False):
  bases = [[]]
  done_bases = []
  # on_length = 1
  
  while len(bases) > 0:
    B = bases[0]
    Bshow = list(B)
    Bshow.sort()
    Bshow.sort(key=len)
    bases = bases[1:]
    print("Evaluating basis: ",list(Bshow),"\t(",len(bases),"more in queue )")
    C = AvClass(B, 6)
    diffs = [i for i in range(1,len(C)) if len(C[i]) != enum[i-1]]
    if len(diffs) == 0:
      for i in range(7,max_len+1):
        C.extend_to_length(i)
        diffs = [i for i in range(1,len(C)) if len(C[i]) != enum[i-1]]
        if len(diffs) > 0:
          break
    if len(diffs) == 0:
      # Now check all the way
      C.extend_to_length(max_len+check_ahead)
      if all([len(C[i]) == enum[i-1] for i in range(1, len(C))]):
        print("\tPerfect match!")
        done_bases.append(B)
    else:
      first_diff = min(diffs)
      if len(C[first_diff]) <= enum[first_diff-1]:
        if verbose:
          print("\tNo chance.");
        continue
      if verbose:
        print("\tCurrent Enum:",[len(C[i]) for i in range(1,len(C))])
      if len(B) == 0:
        to_try = sorted([P for P in C[first_diff] if P.is_representative()])
      elif max([len(b) for b in B]) < first_diff:
        # try all basis elements in lex order
        to_try = sorted(list(C[first_diff]))
        if verbose:
          print("\tPossible new basis elts:",to_try)
      else:
        # try all basis elements after last one (lex) of cur length in B
        to_try = sorted(list(C[first_diff]))
        to_try = [b for b in to_try if b > max([t for t in B if len(t) == first_diff])]
        if verbose:
          print("\tPossible new basis elts:",to_try)
      if len(to_try) == 0:
        # nothing else to try of this length
        continue
      to_try = kill_auts(B, to_try)
      k = len(to_try)
      ci = 0
      for new_b in to_try:
        ci += 1
        new_B = PermSet(B).union(PermSet([new_b]))
        if verbose:
          print("\t\tTrying basis:",sorted(list(new_B)),"\t(",ci,"/",k,")")
        elif ci % 25 == 0:
          print("\t(",ci,"/",k,")")
        new_C = AvClass(new_B, first_diff+check_ahead)
        if all([len(new_C[i]) >= enum[i-1] for i in range(first_diff+1, first_diff+check_ahead+1)]):
          if verbose:
            print("\t\t\tOkay. New enum:",[len(new_C[i]) for i in range(1,len(new_C))])
          
            # print("\t\t\t\tMatches check_ahead: checking all the way.")
            # new_C.extend_to_length(max_len+check_ahead)
            # if all([len(new_C[i]) == enum[i-1] for i in range(1,max_len+check_ahead+1)]):
            #   print("\t\t\t\t\tMatches all the way!")
            #   done_bases.append(new_B)
            #   continue
            # else:
            #   print("\t\t\t\tNot yet a perfect match.")
          bases.append(new_B)
        else:
          if verbose:
            print("\t\t\tToo Small.")
    # Kill syms in list
    old_size = len(bases)
    bases = kill_syms(bases)
    new_size = len(bases)
    if verbose:
      print("Killed off symmetries: ",old_size,"->",new_size)

  return [list(B) for B in done_bases]



# def all_rtlmax_end(P,bound=1):
#   L = P.rtlmax()
#   return L == range(len(P)-len(L), len(P)) and len(L) >= bound

# from collections import defaultdict
# import csv

# ltd = 10

# def prof(P):
#   return tuple(tupify([f[1](P) for f in stats]))

# def tupify(L):
#   S = []                                    
#   for p in L:
#     if isinstance(p,list):
#       S.append(tuple(p))
#     else:
#       S.append(p)
#   return tuple(S)

# def sf(P):
#   return str(P).replace(' ', '')

# def pos_of_least_val_block_inc(P):
#   L = [P.index(0)]
#   v = 1
#   while P.index(v) > P.index(v-1) and v < len(P):
#     L.append(P.index(v))
#     v += 1
#   return L

# def pos_of_greatest_val_block_inc(P):
#   L = [P.index(len(P)-1)]
#   v = len(P)-1
#   while P.index(v) > P.index(v-1):
#     L.append(P.index(v-1))
#     v -= 1
#   return L[::-1]



# stats = [
#   ('position of descents', lambda P : [i for i in range(len(P)-1) if P[i] > P[i+1]]),
#   ('positions of ltrmin', Perm.ltrmin),
#   ('positions of rtlmax', Perm.rtlmax),
#   ('position of + bonds', lambda P : [i for i in range(len(P)-1) if P[i]+1==P[i+1]]),
#   # ('least_val_block', pos_of_least_val_block_inc),
#   # ('greatest_val_block', pos_of_greatest_val_block_inc),
# ]

# av_2143 = defaultdict(set)
# av_3142 = defaultdict(set)

# C = Av([3142], ltd)
# D = Av([2143,415263], ltd)

# for P in C[ltd]:
#   if not P.skew_decomposable() and P not in D[ltd] and all_rtlmax_end(P,2) and P.num_inc_bonds() == 0:
#     av_3142[prof(P)].add(P)

# for P in D[ltd]:
#   if not P.skew_decomposable() and P not in C[ltd] and all_rtlmax_end(P,2) and P.num_inc_bonds() == 0:
#     av_2143[prof(P)].add(P)

# S = set(av_2143.keys()+av_3142.keys())

# with open('data.csv', 'wb') as cf:
#   w = csv.writer(cf)
#   for k in S:
#     w.writerow(list(k)+[' / '.join([sf(P) for P in av_2143[k]])]+[' / '.join([sf(P) for P in av_3142[k]])])




from .permutation import *
from .permset import *
from .permclass import *
import collections, itertools, operator
from math import factorial

class GeometricGridClass(PermClass):
  
  ## Todo: Automatic row/col signs if possible, otherwise use the x2 trick

  def __init__(self, M, col=None, row=None, max_length=8, generate=True): 

    M = M[::-1]
    M = map(list, zip(*M))
    self.M = M
    self.col = col
    self.row = row
    list.__init__(self, [PermSet() for i in range(0, max_length+1)])

    if self.col == None or self.row == None:
      (self.col, self.row) = self.guess_signs(self.M)

    self.alphabet_size = sum([(1 if self.M[i][j] != 0 else 0) for i in range(0,len(self.M)) for j in range(0,len(self.M[i]))])
    self.alphabet = [(i,j) for i in range(0, len(self.M)) for j in range(0,len(self.M[i])) if self.M[i][j] != 0]

    self.commuting_pairs = []

    self.dots = []
    for (index, coords) in enumerate(self.alphabet):
      if self.M[coords[0]][coords[1]] == 2:
        self.dots.append(index)
    
    for (i,l) in enumerate(self.alphabet):
      for j in range(i,len(self.alphabet)):
        if self.alphabet[i][0] != self.alphabet[j][0] and self.alphabet[i][1] != self.alphabet[j][1]:
          self.commuting_pairs.append((i,j))

    self.alphabet_indices = [i for i in range(0,len(self.alphabet))]     

    if generate:
      self.generate_perms(max_length)

    self.length = max_length

  def find_word_for_perm(self, P):
    l = len(P)

    M = self.M
    column_signs = self.col
    row_signs = self.row
    
    all_words = itertools.product(self.alphabet_indices,repeat=l)
      
    for word in all_words:
      perm = self.dig_word_to_perm(word)
      if perm == P:
        return dig_to_num(word)

  def guess_signs(self, M):
    col = [0]*len(M)
    row = [0]*len(M[0])

    for c in range(len(col)):
      kc = 1
      for r in range(len(row)):
        if row[r] != 0 and M[c][r] != 0 and M[c][r] != 2:
          kc = row[r] * M[c][r]
      col[c] = kc

      for r in range(len(row)):
        if M[c][r] != 0 and M[c][r] != 2:
          row[r] = col[c] * M[c][r]
        else:
          row[r] = col[c]

    # print 'col:',col,'\t\trow:',row
    # print M

    # CHECKING
    for c in range(len(col)):
      for r in range(len(row)):
        if M[c][r] != 0 and M[c][r] != 2 and M[c][r] != col[c] * row[r]:
          # print 'c:',c,',  r:',r,',  M:',M[c][r]
          raise Exception("Unable to find row and columns signs. Please enter them manually.")
          # return (False, False)
    return (col, row)
    
  def generate_perms(self, max_length):
    M = self.M
    column_signs = self.col
    row_signs = self.row

    list.__init__(self, [PermSet() for i in range(0, max_length+1)])
    self[1].add(Permutation([1])) 
    
    for length in range(2,max_length+1):
      ''' Try all words of length 'length' with alphabet 
      equal to the cell alphabet of M.'''
      all_words = itertools.product(self.alphabet_indices,repeat=length)
      
      for word in all_words:
        P = self.dig_word_to_perm(word)
        if P:
          self[length].add(P)

  def dig_word_to_perm(self, word, ignore_bad=False):
    if not ignore_bad:
      bad_word = False
      for index in self.dots:
        if word.count(index) > 1:
          return False
      if not self.is_valid_word(word):
        return False
    points = []
    height = len(word)+2
    for (position, letter) in enumerate(word):
      if self.col[self.alphabet[letter][0]] == 1:
        x_point = (self.alphabet[letter][0]+1)*height - position
      else:
        x_point = (self.alphabet[letter][0])*height + position
      if self.row[self.alphabet[letter][1]] == 1:
        y_point = (self.alphabet[letter][1]+1)*height - position
      else:
        y_point = (self.alphabet[letter][1])*height + position
      points.append((x_point,y_point))
    points.sort(key=operator.itemgetter(0))  
    widths = [p[0] for p in points] 
    heights = [p[1] for p in points] 
    new_points = Permutation.standardize(heights)
    return Permutation(new_points)

#   def alpha_word_to_perm(self, word):
#     w = []
#     for i in range(0, len(word)):
#       n = ord(word[i])-97
#       if n < 0 or n >= self.alphabet_size:
#         return False
#       w.append(n)
#     return self.dig_word_to_perm(w, ignore_bad=True)

  def is_valid_word(self, word):
    for i in range(0,len(word)-1):
      if (word[i+1], word[i]) in self.commuting_pairs:
        return False
    return True

#   def find_inflations_avoiders(self, basis):
#     max_basis_length = max([len(B) for B in basis])
#     allowed_inflations = [Permutation([1])]
#     allowed_inflations.extend(list(set([P for sublist in [B.all_intervals(return_patterns=True) for B in basis] for P in sublist])))
#     avoidence_inflations = []
#     print(allowed_inflations)
#     for length in range(2,max_basis_length):
#       all_words = itertools.product(self.alphabet_indices,repeat=length)
#       for word in all_words:
#         P = self.dig_word_to_perm(word)
#         if not P:
#           continue
#         combos = itertools.product(range(0,len(allowed_inflations)),repeat=len(P))
#         for combo in combos:
#           components = [allowed_inflations[i] for i in combo]
#           Q = P.inflate(components)
#           if Q in basis:
#             what_to_avoid = [chr(e+97)+'_'+str(allowed_inflations.index(components[i])) for (i,e) in enumerate(word)]
#             print(what_to_avoid,'=',Q)
#             if what_to_avoid not in avoidence_inflations:
#               avoidence_inflations.append(what_to_avoid)
#     return avoidence_inflations

#   def inflation_rules(self, basis):
#     avoidence_inflations = self.find_inflations_avoiders(basis)
#     rules = []
#     letters = set()
#     for ai in avoidence_inflations:
#       rule = 'SS, '
#       rule += (', SS, '.join(ai))
#       rule += ', SS'
#       rules.append('C(P('+rule+'))')
#       letters = letters.union(ai)
#     big_rule = 'I(' + (','.join(rules)) + ')'
#     letters = list(letters)
#     letters.sort()
#     return (letters, big_rule)

# def dig_to_num(w):
#   return ''.join([chr(a+97) for a in w])








from ..permutation import *
from ..avclass import *





class RestrictedContainer(object):

  def __init__(self, basis_list, input_list, container = [], output_list = []):
    self.input_list = (input_list if type(input_list) == list else range(1,input_list+1))
    self.output_list = output_list
    self.basis_list = [Permutation(b) for b in basis_list]
    self.container = container


  def push(self, position):
    if len(self.input_list) == 0:
      return -1
    new_container = self.container[:position] + [self.input_list[0]] + self.container[position:]
    for basis in self.basis_list:
      if basis.involved_in( Permutation(new_container) ):
        return -1
    else:
      return RestrictedContainer(self.basis_list, self.input_list[1:], new_container, self.output_list)

  def pop(self):
    if len(self.container) == 0:
      return -1
    popped = self.container[0]
    new_output = self.output_list + [popped]
    return RestrictedContainer(self.basis_list, self.input_list, self.container[1:], new_output)

  def __repr__(self):
    return "{} | {} | {}".format(self.output_list,self.container, self.input_list)

  def container_size(self):
    return len(self.container)

  def output_size(self):
    return len(self.input_list)

  def input_size(self):
    return len(self.input_list)

  def output(self):
    return self.output_list











from math import factorial
import types

import permpy.permset
import permpy.permclass
from permpy.permutation import Permutation
from permpy.permset import PermSet


class AvClass(permpy.permclass.PermClass):
    """Object representing an avoidance class
    >>> p = Permutation(123)
    >>> av = AvClass(p)
    """


    def __init__(self, basis, length=8, verbose=0):
        list.__init__(self, [PermSet() for i in range(0, length+1)])
        self.length = length

        temp_basis = []
        for P in basis:
            temp_basis.append(Permutation(P))
        basis = temp_basis
        self.basis = basis

        if length >= 1:
                self[1].add(Permutation([1]));
        for n in range(2,length+1):
            k = 0
            outof = len(self[n-1])
            for P in self[n-1]:
                k += 1
                if verbose > 0 and k % verbose == 0:
                    # print '\t\t\t\tRight Extensions:',k,'/',outof,'\t( length',n,')'
                    print('\t\t\t\tRight Extenstions: {}/{}\t( length {}'.format(
                                k, outof, n))
                insertion_locations = P.insertion_locations
                add_this_time = []
                for Q in P.right_extensions():
                    is_good = True
                    for B in basis:
                        if B.involved_in(Q,last_require=2):
                            is_good = False
                            insertion_locations[Q[-1]] = 0
                            # break
                    if is_good:

                        add_this_time.append(Q)
                for Q in add_this_time:
                    # print Q,'is good'
                    # print '\tchanging IL from ',Q.insertion_locations,'to',(insertion_locations[:Q[-1]+1]+    insertion_locations[Q[-1]:])
                    Q.insertion_locations = insertion_locations[:Q[-1]+1] + insertion_locations[Q[-1]:]
                    self[n].add(Q)

    def extend_to_length(self, l):
        for i in range(self.length+1, l+1):
            self.append(permset.PermSet())
        if (l <= self.length):
            return
        old = self.length
        self.length = l
        for n in range(old+1,l+1):
            for P in self[n-1]:
                insertion_locations = P.insertion_locations
                add_this_time = []
                for Q in P.right_extensions():
                    is_good = True
                    for B in self.basis:
                        if B.involved_in(Q,last_require=2):
                            is_good = False
                            insertion_locations[Q[-1]] = 0
                            # break
                    if is_good:

                        add_this_time.append(Q)
                for Q in add_this_time:
                    # print Q,'is good'
                    # print '\tchanging IL from ',Q.insertion_locations,'to',(insertion_locations[:Q[-1]+1]+    insertion_locations[Q[-1]:])
                    Q.insertion_locations = insertion_locations[:Q[-1]+1] + insertion_locations[Q[-1]:]
                    self[n].add(Q)

    def right_juxtaposition(self, C, generate_perms=True):
        A = permset.PermSet()
        max_length = max([len(P) for P in self.basis]) + max([len(P) for P in C.basis])
        for n in range(2, max_length+1):
            for i in range(0, factorial(n)):
                P = Permutation(i,n)
                for Q in self.basis:
                    for R in C.basis:
                        if len(Q) + len(R) == n:
                            if (Q == Permutation(P[0:len(Q)]) and R == Permutation(P[len(Q):n])):
                                A.add(P)
                        elif len(Q) + len(R) - 1 == n:
                            if (Q == Permutation(P[0:len(Q)]) and Permutation(R) == Permutation(P[len(Q)-1:n])):
                                A.add(P)
        return AvClass(list(A.minimal_elements()), length=(8 if generate_perms else 0))

    def above_juxtaposition(self, C, generate_perms=True):
        inverse_class = AvClass([P.inverse() for P in C.basis], 0)
        horizontal_juxtaposition = self.right_juxtaposition(inverse_class, generate_perms=False)
        return AvClass([B.inverse() for B in horizontal_juxtaposition.basis], length=(8 if generate_perms else 0))

    def contains(self, C):
        for P in self.basis:
            good = False
            for Q in C.basis:
                if P.involved_in(Q):
                    good = True
                    break
            if not good:
                return False
        return True




from __future__ import print_function
import sys, os, subprocess, time
import math
import random
import fractions
import itertools
from math import factorial
import types

import copy
import time
from math import factorial


__author__ = 'Cheyne Homberger, Jay Pantone'



# a class for creating permutation objects
class Permutation(tuple):
  'can be initialized with either (index,length) or a list of entries'

  # static class variable, controls permutation representation
  _REPR = 'oneline'
  lower_bound = []
  upper_bound = []
  bounds_set = False;
  insertion_locations = []

  #===============================================================%
  # some useful functions for playing with permutations

  @staticmethod
  def monotone_increasing(n):
    return Permutation(range(n))

  @staticmethod
  def monotone_decreasing(n):
    return Permutation(range(n)[::-1])

  @staticmethod
  def random(n):
    '''outputs a random permutation of length n'''
    L = range(n)
    random.shuffle(L)
    return Permutation(L)

  @staticmethod
  def random_avoider(n, B, simple=False, involution=False, verbose=-1):
    i = 1
    p = Permutation.random(n)
    while (involution and not p.is_involution()) or (simple and not p.is_simple()) or not p.avoids_set(B):
      i += 1
      p = Permutation.random(n)
      if verbose != -1 and i % verbose == 0:
        print("Tested: "+str(i)+" permutations.");
    return p


  @staticmethod
  def listall(n):
    '''returns a list of all permutations of length n'''
    if n == 0:
      return []
    else:
      L = []
      for k in range(math.factorial(n)):
        L.append(Permutation(k,n))
      return L

  @staticmethod
  def standardize(L):
    # copy by value
    assert len(set(L)) == len(L), 'make sure elements are distinct!'
    ordered = L[:]
    ordered.sort()
    return [ordered.index(x) for x in L]

  @staticmethod
  def change_repr():
    '''changes between cycle notation or one-line notation. note that
    internal representation is still one-line'''
    L = ['oneline', 'cycles', 'both']
    k = input('1 for oneline, 2 for cycles, 3 for both\n ')
    k -= 1
    Permutation._REPR = L[k]

  @staticmethod
  def ind2perm(k, n):
    '''de-indexes a permutation'''
    result = list(range(n))
    def swap(i,j):
      t = result[i]
      result[i] = result[j]
      result[j] = t
    for i in range(n, 0, -1):
      j = k % i
      swap(i-1,j)
      k //= i
    return Permutation(result)

  #================================================================#
  # overloaded built in functions:

  def __new__(cls, p, n = None):
    ''' Initializes a permutation object, internal indexing starts at zero. '''
    entries = []
    if n:
      return Permutation.ind2perm(p, n)
    else:
      if isinstance(p, Permutation):
        return tuple.__new__(cls, p)
      elif isinstance(p, tuple):
        entries = list(p)[:]
      elif isinstance(p, list):
        entries = p[:]
      elif isinstance(p, int):
        p = str(p)
        entries = list(p)
      # standardizes, starting at zero
      assert len(set(entries)) == len(entries), 'elements not distinct'
      assert len(entries) > 0 or p==list() or p==set() or p==tuple(), 'invalid permutation'
      entries.sort()
      standardization =  map(lambda e: entries.index(e), p)
      return tuple.__new__(cls, standardization)

  def __init__(self,p,n=None):
    self.insertion_locations = [1]*(len(self)+1)

  def __call__(self,i):
    '''allows permutations to be used as functions
    (useful for counting cycles)'''
    return self[i]

  def oneline(self):
    s = ' '
    for i in self:
      s += str( i+1 ) + ' '
    return s

  def cycles(self):
    stringlist = ['( ' + ' '.join([str(x+1) for x in cyc]) + ' )'
                                    for cyc in self.cycle_decomp()]
    return ' '.join(stringlist)

  def __repr__(self):
    '''tells python how to display a permutation object'''
    if Permutation._REPR == 'oneline':
      return self.oneline()
    if Permutation._REPR == 'cycles':
      return self.cycles()
    else:
      return '\n '.join([self.oneline(), self.cycles()])

  # def __hash__(self):
  #   '''allows fast comparisons of permutations, and allows sets of
  #   permutations'''
  #   return (self.perm2ind(), self.__len__()).__hash__()

  # def __eq__(self,other):
  #   ''' checks if two permutations are equal '''
  #   if len(self) != len(other):
  #     return False
  #   for i in range(len(self)):
  #     if self[i] != other[i]:
  #       return False
  #   return True

  # def __ne__(self,other):
  #   return not self == other

  def __mul__(self,other):
    ''' multiplies two permutations '''
    assert len(self) == len(other)
    L = list(other)
    for i in range(len(L)):
      L[i] = self.__call__(L[i])
    return Permutation(L)

  def __pow__(self, power):
    assert isinstance(power, int) and power >= 0
    if power == 0:
      return Permutation(range(len(self)))
    else:
      ans = self
      for i in range(power - 1):
        ans *= self
      return ans

  def perm2ind(self):
    ''' De-indexes a permutation. '''
    q = list(self)
    n = self.__len__()
    def swap(i,j):
      t = q[i]
      q[i] = q[j]
      q[j] = t
    result = 0
    multiplier = 1
    for i in range(n-1,0,-1):
      result += q[i]*multiplier
      multiplier *= i+1
      swap(i, q.index(i))
    return result

  def delete(self,i):
    p = list(self)
    del p[i]
    return Permutation(p)

  def ins(self,i,j):
    p = list(self)
    for k in range(len(p)):
      if p[k] >= j:
        p[k] += 1
    p = p[:i] + [j] + p[i:]
    return Permutation(p)

  # returns the complement of the permutation
  def complement(self):
    n = self.__len__()
    L = [n-1-i for i in self]
    return Permutation(L)

  # returns the reverse of the permutation
  def reverse(self):
    q = list(self)
    q.reverse()
    return Permutation(q)

  def inverse(self):
    p = list(self)
    n = self.__len__()
    q = [0 for j in range(n)]
    for i in range(n):
      q[p[i]] = i
    return Permutation(q)

  def plot(self):
    ''' Draws a plot of the given Permutation. '''
    n = self.__len__()
    array = [[' ' for i in range(n)] for j in range(n)]
    for i in range(n):
      array[self[i]][i] = '*'
    array.reverse()
    s = '\n'.join( (''.join(l) for l in array))
    # return s
    print(s)

  def cycle_decomp(self):
    n = self.__len__()
    seen = set()
    cyclelist = []
    while len(seen) < n:
      a = max(set(range(n)) - seen)
      cyc = [a]
      b = self(a)
      seen.add(b)
      while b != a:
        cyc.append(b)
        b = self(b)
        seen.add(b)
      cyclelist.append(cyc)
    cyclelist.reverse()
    return cyclelist

  def num_disjoint_cycles(self):
    return len(self.cycle_decomp())

  def sum(self, Q):
    return Permutation(list(self)+[i+len(self) for i in Q])

  def skew_sum(self, Q):
    return Permutation([i+len(Q) for i in self]+list(Q))


# Permutation Statistics - somewhat self-explanatory

  def fixed_points(self):
    sum = 0
    for i in range(self.__len__()):
      if self(i) == i:
        sum += 1
    return sum


  def skew_decomposable(self):
    p = list(self)
    n = self.__len__()
    for i in range(1,n):
      if set(range(n-i,n)) == set(p[0:i]):
        return True
    return False

  def sum_decomposable(self):
    p = list(self)
    n = self.__len__()
    for i in range(1,n):
      if set(range(0,i)) == set(p[0:i]):
        return True
    return False


  def numcycles(self):
    num = 1
    n = self.__len__()
    list = range(n)
    k = 0
    while list:
      if k in list:
        del list[list.index(k)]
        k = self(k)
      else:
        k = list[0]
        num += 1
    return num

  def descents(self):
    p = list(self)
    n = self.__len__()
    numd = 0
    for i in range(1,n):
      if p[i-1] > p[i]:
        numd+=1
    return numd

  def ascents(self):
    return self.__len__()-1-self.descents()

  def bends(self):
        # Bends measures the number of times the permutation p
        # "changes direction".  Bends is also the number of
        # non-monotone consecutive triples in p.
        # The permutation p can be expressed as the concatenation of
        # bend(p)+1 monotone segments.
    return len([i for i in range(1, len(self)-1) if (self[i-1] > self[i] and self[i+1] > self[i]) or (self[i-1] < self[i] and self[i+1] < self[i])])

  def peaks(self):
    return len([i for i in range(1, len(self)-1) if self[i-1] < self[i] and self[i+1] < self[i]])

  def trivial(self):
    return 0

  def order(self):
    L = map(len, self.cycle_decomp())
    return reduce(lambda x,y: x*y / fractions.gcd(x,y), L)

  def ltrmin(self):
    p = list(self)
    n = self.__len__()
    L = []
    for i in range(n):
      flag = True
      for j in range(i):
        if p[i] > p[j]:
          flag = False
      if flag: L.append(i)
    return L

  def rtlmin(self):
    return [len(self)-i-1 for i in Permutation(self[::-1]).ltrmin()]

  def ltrmax(self):
    return [len(self)-i-1 for i in Permutation(self[::-1]).rtlmax()][::-1]

  def rtlmax(self):
    return [len(self)-i-1 for i in self.complement().reverse().ltrmin()][::-1]

  def numltrmin(self):
    p = list(self)
    n = self.__len__()
    num = 1
    m = p[0]
    for i in range(1,n):
      if p[i] < m:
        num += 1
        m = p[i]
    return num

  def inversions(self):
    p = list(self)
    n = self.__len__()
    inv = 0
    for i in range(n):
      for j in range(i+1,n):
        if p[i]>p[j]:
          inv+=1
    return inv

  def noninversions(self):
    p = list(self)
    n = self.__len__()
    inv = 0
    for i in range(n):
      for j in range(i+1,n):
        if p[i]<p[j]:
          inv+=1
    return inv

  def bonds(self):
    numbonds = 0
    p = list(self)
    for i in range(1,len(p)):
      if p[i] - p[i-1] == 1 or p[i] - p[i-1] == -1:
        numbonds+=1
    return numbonds

  def majorindex(self):
    sum = 0
    p = list(self)
    n = self.__len__()
    for i in range(0,n-1):
      if p[i] > p[i+1]:
        sum += i + 1
    return sum

  def fixedptsplusbonds(self):
    return self.fixed_points() + self.bonds()

  def longestrunA(self):
    p = list(self)
    n = self.__len__()
    maxi = 0
    length = 1
    for i in range(1,n):
      if p[i-1] < p[i]:
        length += 1
        if length > maxi: maxi = length
      else:
        length = 1
    return max(maxi,length)

  def longestrunD(self):
    return self.complement().longestrunA()

  def longestrun(self):
    return max(self.longestrunA(),self.longestrunD())

  def christiecycles(self):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product. used for transpositions
    p = list(self)
    n = self.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = range(1,n+1) + [0]
    blackperm = [0 for i in range(n+1)]
    for i in range(n+1):
      ind = q.index(i)
      blackperm[i] = q[ind-1]
    newperm = []
    for i in range(n+1):
      k = blackperm[i]
      j = grayperm[k]
      newperm.append(j)
    return Permutation(newperm).numcycles()

  def othercycles(self):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product
    p = list(self)
    n = self.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = [n] + range(n)
    blackperm = [0 for i in range(n+1)]
    for i in range(n+1):
      ind = q.index(i)
      blackperm[i] = q[ind-1]
    newperm = []
    for i in range(n+1):
      k = blackperm[i]
      j = grayperm[k]
      newperm.append(j)
    return Permutation(newperm).numcycles()

  def sumcycles(self):
    return self.othercycles() + self.christiecycles()

  def maxcycles(self):
    return max(self.othercycles() - 1,self.christiecycles())

  def is_involution(self):
    return self == self.inverse()

  def threepats(self):
    p = list(self)
    n = self.__len__()
    patnums = {'123' : 0, '132' : 0, '213' : 0,
               '231' : 0, '312' : 0, '321' : 0}
    for i in range(n-2):
      for j in range(i+1,n-1):
        for k in range(j+1,n):
          patnums[''.join(map(lambda x:
                              str(x+1),Permutation([p[i], p[j], p[k]])))] += 1
    return patnums

  def fourpats(self):
    p = list(self)
    n = self.__len__()
    patnums = {'1234' : 0, '1243' : 0, '1324' : 0,
               '1342' : 0, '1423' : 0, '1432' : 0,
               '2134' : 0, '2143' : 0, '2314' : 0,
               '2341' : 0, '2413' : 0, '2431' : 0,
               '3124' : 0, '3142' : 0, '3214' : 0,
               '3241' : 0, '3412' : 0, '3421' : 0,
               '4123' : 0, '4132' : 0, '4213' : 0,
               '4231' : 0, '4312' : 0, '4321' : 0 }

    for i in range(n-3):
      for j in range(i+1,n-2):
        for k in range(j+1,n-1):
          for m in range(k+1,n):
            patnums[''.join(map(lambda x:
                      str(x+1),list(Permutation([p[i], p[j], p[k], p[m]]))))] += 1
    return patnums

  def num_consecutive_3214(self):
    number = 0
    n = len(self)
    for i in range(n-3):
      if self[i+2] < self[i+1] < self[i] < self[i+3]:
        number += 1
    return number

  def coveredby(self):
    S = set()
    n = len(self)
    for i in range(n+1):
      for j in range(n+1):
        S.add(self.ins(i,j))
    return S

  def buildupset(self, height):
    n = len(self)
    L = [set() for i in range(n)]
    L.append( set([self]) )
    for i in range(n + 1, height):
      oldS = list(L[i-1])
      newS  = set()
      for perm in oldS:
        newS = newS.union(perm.coveredby())
      L.append(newS)
    return L

  def set_up_bounds(self):
    L = list(self)
    n = len(L)
    upper_bound = [-1]*n
    lower_bound = [-1]*n
    for i in range(0,n):
      min_above = -1
      max_below = -1
      for j in range(i+1,len(self)):
        if L[j] < L[i]:
          if L[j] > max_below:
            max_below = L[j]
            lower_bound[i] = j
        else:
          if L[j] < min_above or min_above == -1:
            min_above = L[j]
            upper_bound[i] = j
    return (lower_bound, upper_bound)

  def avoids(self, P,lr=0):
    return not P.involved_in(self, last_require=lr)

  def avoids_set(self, B):
    for b in B:
      if b.involved_in(self):
        return False
    return True

  def involves(self, P, lr=0):
    return P.involved_in(self,last_require=lr)

  def involved_in(self, P, last_require=0):
    if not self.bounds_set:
      (self.lower_bound, self.upper_bound) = self.set_up_bounds()
      self.bounds_set = True
    L = list(self)
    n = len(L)
    p = len(P)
    if n <= 1 and n <= p:
      return True

    indices = [0]*n

    if last_require == 0:
      indices[len(self)-1] = p - 1
      while indices[len(self)-1] >= 0:
        if self.involvement_check(self.upper_bound, self.lower_bound, indices, P, len(self)-2):
          return True
        indices[len(self)-1] -= 1
      return False
    else:
      for i in range(1, last_require+1):
        indices[n-i] = p-i
      if not self.involvement_check_final(self.upper_bound, self.lower_bound, indices, P, last_require):
        return False

      return self.involvement_check(self.upper_bound, self.lower_bound, indices, P, len(self) - last_require - 1)

  def involvement_check_final(self, upper_bound, lower_bound, indices, q, last_require):
    for i in range(1,last_require):
      if not self.involvement_fits(upper_bound, lower_bound, indices, q, len(self)-i-1):
        return False
    return True

  def involvement_check(self, upper_bound, lower_bound, indices, q, next):
    if next < 0:
      return True
    # print indices,next
    indices[next] = indices[next+1]-1

    while indices[next] >= 0:
      if self.involvement_fits(upper_bound, lower_bound, indices, q, next) and self.involvement_check(upper_bound, lower_bound, indices, q, next-1):
        return True
      indices[next] -= 1
    return False

  def involvement_fits(self, upper_bound, lower_bound, indices, q, next):
    return (lower_bound[next] == -1 or q[indices[next]] > q[indices[lower_bound[next]]]) and (upper_bound[next] == -1 or q[indices[next]] < q[indices[upper_bound[next]]])


  def occurrences(self, pattern):
    total = 0
    for subseq in itertools.combinations(self, len(pattern)):
      if Permutation(subseq) == pattern:
        total += 1
    return total



  def all_intervals(self, return_patterns=False):
    blocks = [[],[]]
    for i in range(2, len(self)):
      blocks.append([])
      for j in range (0,len(self)-i+1):
        if max(self[j:j+i]) - min(self[j:j+i]) == i-1:
          blocks[i].append(j)
    if return_patterns:
      patterns = []
      for length in range(0, len(blocks)):
        for start_index in blocks[length]:
          patterns.append(Permutation(self[start_index:start_index+length]))
      return patterns
    else:
      return blocks



  def all_monotone_intervals(self):
    mi = []
    difference = 0
    c_start = 0
    c_length = 0
    for i in range(0,len(self)-1):
      if math.fabs(self[i] - self[i+1]) == 1 and (c_length == 0 or self[i] - self[i+1] == difference):
        if c_length == 0:
          c_start = i
        c_length += 1
        difference = self[i] - self[i+1]
      else:
        if c_length != 0:
          mi.append((c_start, c_start+c_length))
        c_start = 0
        c_length = 0
        difference = 0
    if c_length != 0:
      mi.append((c_start, c_start+c_length))
    return mi

  def maximal_interval(self):
    ''' finds the biggest interval, and returns (i,j) is one is found,
			where i is the size of the interval, and j is the index
			of the first entry in the interval

		returns (0,0) if no interval is found, i.e., if the permutation
			is simple'''
    for i in range(2, len(self))[::-1]:
      for j in range (0,len(self)-i+1):
        if max(self[j:j+i]) - min(self[j:j+i]) == i-1:
          return (i,j)
    return (0,0)

  def simple_location(self):
    ''' searches for an interval, and returns (i,j) if one is found,
			where i is the size of the interval, and j is the
			first index of the interval

		returns (0,0) if no interval is found, i.e., if the permutation
			is simple'''
    mins = list(self)
    maxs = list(self)
    for i in range(1,len(self)-1):
      for j in reversed(range(i,len(self))):
        mins[j] = min(mins[j-1], self[j])
        maxs[j] = max(maxs[j-1], self[j])
        if maxs[j] - mins[j] == i:
          return (i,j)
    return (0,0)

  def is_simple(self):
    ''' returns True is this permutation is simple, False otherwise'''
    (i,j) = self.simple_location()
    return i == 0

  def is_strongly_simple(self):
    return self.is_simple() and all([p.is_simple() for p in self.children()])

  def decomposition(self):
    base = Permutation(self)
    components = [Permutation([1])for i in range(0,len(base))]
    while not base.is_simple():
      assert len(base) == len(components)
      (i,j) = base.maximal_interval()
      assert i != 0
      interval = list(base[j:i+j])
      new_base = list(base[0:j])
      new_base.append(base[j])
      new_base.extend(base[i+j:len(base)])
      new_components = components[0:j]
      new_components.append(Permutation(interval))
      new_components.extend(components[i+j:len(base)])
      base = Permutation(new_base)
      components = new_components
    return (base, components)

  def inflate(self, components):
    assert len(self) == len(components), 'number of components must equal length of base'
    L = list(self)
    NL = list(self)
    current_entry = 0
    for entry in range(0, len(self)):
      index = L.index(entry)
      NL[index] = [components[index][i]+current_entry for i in range(0, len(components[index]))]
      current_entry += len(components[index])
    NL_flat = [a for sl in NL for a in sl]
    return Permutation(NL_flat)

  def right_extensions(self):
    L = []
    if len(self.insertion_locations) > 0:
      indices = self.insertion_locations
    else:
      indices = [1]*(len(self)+1)

    R = [j for j in range(len(indices)) if indices[j] == 1]
    for i in R:
      j = 0
      A = [self[j] + (1 if self[j] > i-1 else 0) for j in range(0,len(self))]
      A.append(i)
      L.append(Permutation(A))
    return L

  # def all_right_extensions(self, max_length, l, S):
  #   if l == max_length:
  #     return S
  #   else:
  #     re = self.right_extensions()
  #     for p in re:
  #       S.add(p)
  #       S = p.all_right_extensions(max_length, l+1, S)
  #   return S

  def all_extensions(self):
    S = set()
    for i in range(0, len(self)+1):
      for j in range(0, len(self)+1):
        # insert (i-0.5) after entry j (i.e., first when j=0)
        l = list(self[:])
        l.insert(j, i-0.5)
        S.add(Permutation(l))
    return PermSet(S)

  def show(self):
    if sys.platform == 'linux2':
      opencmd = 'gnome-open'
    else:
      opencmd = 'open'
    s = "\\documentclass{standalone}\n\\usepackage{tikz}\n\n\\begin{document}\n\n"
    s += self.to_tikz()
    s += "\n\n\end{document}"
    dname = random.randint(1000,9999)
    os.system('mkdir t_'+str(dname))
    with open('t_'+str(dname)+'/t.tex', 'w') as f:
      f.write(s)
    subprocess.call(['pdflatex', '-output-directory=t_'+str(dname), 't_'+str(dname)+'/t.tex'],
      stderr = subprocess.PIPE, stdout = subprocess.PIPE)
    # os.system('pdflatex -output-directory=t_'+str(dname)+' t_'+str(dname)+'/t.tex')
    subprocess.call([opencmd, 't_'+str(dname)+'/t.pdf'],
      stderr = subprocess.PIPE, stdout = subprocess.PIPE)
    time.sleep(1)
    if sys.platform != 'linux2':
      subprocess.call(['rm', '-r', 't_'+str(dname)+'/'])

  def to_tikz(self):
    s = r'\begin{tikzpicture}[scale=.3,baseline=(current bounding box.center)]';
    s += '\n\t'
    s += r'\draw[ultra thick] (1,0) -- ('+str(len(self))+',0);'
    s += '\n\t'
    s += r'\draw[ultra thick] (0,1) -- (0,'+str(len(self))+');'
    s += '\n\t'
    s += r'\foreach \x in {1,...,'+str(len(self))+'} {'
    s += '\n\t\t'
    s += r'\draw[thick] (\x,.09)--(\x,-.5);'
    s += '\n\t\t'
    s += r'\draw[thick] (.09,\x)--(-.5,\x);'
    s += '\n\t'
    s += r'}'
    for (i,e) in enumerate(self):
      s += '\n\t'
      s += r'\draw[fill=black] ('+str(i+1)+','+str(e+1)+') circle (5pt);'
    s += '\n'
    s += r'\end{tikzpicture}'
    return s

  def shrink_by_one(self):
    return PermSet([Permutation(p) for p in [self[:i]+self[i+1:] for i in range(0,len(self))]])

  def children(self):
    return self.shrink_by_one()

  def downset(self):
    return PermSet([self]).downset()

  def sum_indecomposable_sequence(self):
    S = self.downset()
    return [len([p for p in S if len(p)==i and not p.sum_decomposable()]) for i in range(1,max([len(p) for p in S])+1)]

  def sum_indec_bdd_by(self, n):
    l = [1]
    S = list(self.children())
    while len(S) > 0 and len(S[0]) > 0:
      l = [len([s for s in S if not s.sum_decomposable()])]+l
      if l[0] > n:
        return False
      S = list(PermSet(S).layer_down())
    return True

  def contains_locations(self, Q):
    locs = []
    sublocs = itertools.combinations(range(len(self)), len(Q))
    for subloc in sublocs:
      if Permutation([self[i] for i in subloc]) == Q:
        locs.append(subloc)

    return locs

  def rank_val(self, i):
    return len([j for j in range(i+1,len(self)) if self[j] < self[i]])

  def rank_encoding(self):
    return [self.rank_val(i) for i in range(len(self))]

  def num_rtlmax_ltrmin_layers(self):
    return len(self.rtlmax_ltrmin_decomposition())

  def rtlmax_ltrmin_decomposition(self):
    P = Permutation(self)
    num_layers = 0
    layers = []
    while len(P) > 0:
      num_layers += 1
      positions = sorted(list(set(P.rtlmax()+P.ltrmin())))
      layers.append(positions)
      P = Permutation([P[i] for i in range(len(P)) if i not in positions])
    return layers

  def num_inc_bonds(self):
    return len([i for i in range(len(self)-1) if self[i+1] == self[i]+1])

  def num_dec_bonds(self):
    return len([i for i in range(len(self)-1) if self[i+1] == self[i]-1])

  def num_bonds(self):
    return len([i for i in range(len(self)-1) if self[i+1] == self[i]+1 or self[i+1] == self[i]-1])

  def contract_inc_bonds(self):
    P = Permutation(self)
    while P.num_inc_bonds() > 0:
      for i in range(0,len(P)-1):
        if P[i+1] == P[i]+1:
          P = Permutation(P[:i]+P[i+1:])
          break
    return P

  def contract_dec_bonds(self):
    P = Permutation(self)
    while P.num_dec_bonds() > 0:
      for i in range(0,len(P)-1):
        if P[i+1] == P[i]-1:
          P = Permutation(P[:i]+P[i+1:])
          break
    return P

  def contract_bonds(self):
    P = Permutation(self)
    while P.num_bonds() > 0:
      for i in range(0,len(P)-1):
        if P[i+1] == P[i]+1 or P[i+1] == P[i]-1:
          P = Permutation(P[:i]+P[i+1:])
          break
    return P

  def all_syms(self):
    S = PermSet([self])
    S = S.union(PermSet([P.reverse() for P in S]))
    S = S.union(PermSet([P.complement() for P in S]))
    S = S.union(PermSet([P.inverse() for P in S]))
    return S

  def is_representative(self):
    return self == sorted(self.all_syms())[0]



class PermClass(list):

  # def __init__(self, n = 8):
    # list.__init__(self, [PermSet(permutation.Permutation.listall(i)) for i in range(n + 1)])
    # self.avoids = []
    # self.length = n

  # def __len__(self):
  #   return self.length

  @staticmethod
  def class_from_test(test, l=8, has_all_syms=False):
    C = [PermSet([Permutation([])])]
    for cur_length in range(1,l+1):
      this_len = PermSet([])
      if len(C[cur_length-1]) == 0:
        return PermClass(C)
      to_check = PermSet(set.union(*[P.all_extensions() for P in C[cur_length-1]]))
      to_check = [P for P in to_check if PermSet(P.children()).issubset(C[cur_length-1])]
      while len(to_check) > 0:
        P = to_check.pop()
        if has_all_syms:
          syms = PermSet([
              P,
              P.reverse(),
              P.complement(),
              P.reverse().complement(),
              P.inverse(),
              P.reverse().inverse(),
              P.complement().inverse(),
              P.reverse().complement().inverse()
            ])
        if test(P):
          if has_all_syms:
            for Q in syms:
              this_len.add(Q)
          else:
            this_len.add(P)
        if has_all_syms:
          for Q in syms:
            if Q in to_check:
              to_check.remove(Q)

      C.append(this_len)
    return PermClass(C)



  def filter_by(self, test):
    for i in range(0, len(self)):
      D = list(self[i])
      for P in D:
        if not test(P):
          self[i].remove(P)

  def guess_basis(self, max_length=6, search_mode=False):
    """
      Guess a basis for the class up to "max_length" by iteratively generating
      the class with basis elements known so far ({}, to start with) and adding
      elements which should be avoided to the basis.

      Search mode goes up to the max length in the class and prints out the number
      of basis elements of each length on the way.
    """

    t = time.time()

    assert max_length < len(self), 'class not big enough to check that far'

    if search_mode:
      max_length = len(self)-1

    # Find the first length at which perms are missing.
    not_all_perms = [i for i in range(len(self)) if i >= 1 and len(self[i]) != factorial(i)]

    # If no perms are missing, we have all perms, so return empty basis.
    if len(not_all_perms) == 0:
      return PermSet([])

    # Add missing perms of minimum length to basis.
    start_length = min(not_all_perms)
    basis = PermSet(Permutation.listall(start_length)).difference(self[start_length])

    if search_mode:
      print('\t'+str(len(basis))+' basis elements of length '+str(start_length)+'\t\t'+("{0:.2f}".format(time.time() - t)) + ' seconds')
      t = time.time()

    basis_elements_so_far = len(basis)

    current_length = start_length + 1

    # Go up in length, adding missing perms at each step.
    while current_length <= max_length:
      C = avclass.AvClass(basis, current_length)
      basis = basis.union(C[-1].difference(self[current_length]))

      if search_mode:
        print('\t'+str(len(basis)-basis_elements_so_far)+' basis elements of length ' + str(current_length) + '\t\t' + ("{0:.2f}".format(time.time() - t)) + ' seconds')
        t = time.time()

      basis_elements_so_far = len(basis)

      current_length += 1

    return basis


  # def guess_basis(self, max_length=8):
  #   max_length = min(max_length, len(self)-1)
  #   B = PermSet()
  #   B.update(self.check_tree_basis(max_length, Permutation([1,2]), PermSet()))
  #   B.update(self.check_tree_basis(max_length, Permutation([2,1]), PermSet()))
  #   return B.minimal_elements()

  # def check_tree_basis(self, max_length, R, S):
  #   if R not in self[len(R)]:
  #     for s in S:
  #       if s.involved_in(R):
  #         return S
  #     S.add(R)
  #     return S
  #   else:
  #     if len(R) == max_length:
  #       return S
  #     re = R.right_extensions()
  #     for p in re:
  #       S = self.check_tree_basis(max_length, p, S)
  #     return S

  def plus_class(self,t):
    C = copy.deepcopy(self)
    for i in range(0,t):
      C = C.plus_one_class()
    return C

  def plus_one_class(self):
    D = copy.deepcopy(self)
    D.append(PermSet())
    for l in range(0,len(self)):
      for P in self[l]:
        D[l+1] = D[l+1].union(P.all_extensions())
    return D

class PermSet(set):
  ''' Provides functions for dealing with sets of Permutation objects '''

  def __repr__(self):
    # if len(self) > 10:
    return 'Set of %d permutations' % len(self)
    # else:
      # return set.__repr__(self)

  @staticmethod
  def all(n):
    ''' builds the set of all permutations of length n'''
    return PermSet(Permutation.listall(n))

  def show_all(self):
    return set.__repr__(self)

  def minimal_elements(self):
    B = list(self)
    B = sorted(B, key=len)
    C = B[:]
    n = len(B)
    for (i,b) in enumerate(B):
      # if i % 1 == 0:
        # print i,'/',n
      if b not in C:
        continue
      for j in range(i+1,n):
        if B[j] not in C:
          continue
        if b.involved_in(B[j]):
          C.remove(B[j])
    return PermSet(C)

  def all_syms(self):
    sym_set = [frozenset(self)]
    sym_set.append(frozenset([i.reverse() for i in self]))
    sym_set.append(frozenset([i.complement() for i in self]))
    sym_set.append(frozenset([i.reverse().complement() for i in self]))
    sym_set.extend([frozenset([k.inverse() for k in L]) for L in sym_set])
    return frozenset(sym_set)

  def layer_down(self):
    S = PermSet()
    i = 1
    n = len(self)
    for P in self:
      # if i % 10000 == 0:
        # print('\t',i,'of',n,'. Now with',len(S),'.')
      S.update(P.shrink_by_one())
      i += 1
    return S

  def downset(self, return_class=False):
    bottom_edge = PermSet()
    bottom_edge.update(self)

    done = PermSet(bottom_edge)
    while len(bottom_edge) > 0:
      oldsize = len(done)
      next_layer = bottom_edge.layer_down()
      done.update(next_layer)
      del bottom_edge
      bottom_edge = next_layer
      del next_layer
      newsize = len(done)
      # print '\t\tDownset currently has',newsize,'permutations, added',(newsize-oldsize),'in the last run.'
    if not return_class:
      return done
    cl = [PermSet([])]
    max_length = max([len(P) for P in done])
    for i in range(1,max_length+1):
      cl.append(PermSet([P for P in done if len(P) == i]))
    return permclass.PermClass(cl)


  def total_statistic(self, statistic):
    return sum([statistic(p) for p in self])

  def threepats(self):
    patnums = {'123' : 0, '132' : 0, '213' : 0,
               '231' : 0, '312' : 0, '321' : 0}
    L = list(self)
    for p in L:
      n = len(p)
      for i in range(n-2):
        for j in range(i+1,n-1):
          for k in range(j+1,n):
            std = Permutation.standardize([p[i], p[j], p[k]])
            patnums[''.join([str(x + 1) for x in std])] += 1
    return patnums

  def fourpats(self):
    patnums = {'1234' : 0, '1243' : 0, '1324' : 0,
               '1342' : 0, '1423' : 0, '1432' : 0,
               '2134' : 0, '2143' : 0, '2314' : 0,
               '2341' : 0, '2413' : 0, '2431' : 0,
               '3124' : 0, '3142' : 0, '3214' : 0,
               '3241' : 0, '3412' : 0, '3421' : 0,
               '4123' : 0, '4132' : 0, '4213' : 0,
               '4231' : 0, '4312' : 0, '4321' : 0 }
    L = list(self)
    for p in L:
      n = len(p)
      for i in range(n-3):
        for j in range(i+1,n-2):
          for k in range(j+1,n-1):
            for m in range(k+1,n):
              std = Permutation.standardize([p[i], p[j], p[k], p[m]])
              patnums[''.join([str(x + 1) for x in std])] += 1
    return patnums



class AvClass(PermClass):

  def __init__(self, basis, length=8, verbose=0):
    list.__init__(self, [PermSet() for i in range(0, length+1)])
    self.length = length

    temp_basis = []
    for P in basis:
      temp_basis.append(Permutation(P))
    basis = temp_basis
    self.basis = basis

    if length >= 1:
        self[1].add(Permutation([1]));
    for n in range(2,length+1):
      k = 0
      outof = len(self[n-1])
      for P in self[n-1]:
        k += 1
        if verbose > 0 and k % verbose == 0:
          print('\t\t\t\tRight Extensions:',k,'/',outof,'\t( length',n,')')
        insertion_locations = P.insertion_locations
        add_this_time = []
        for Q in P.right_extensions():
          is_good = True
          for B in basis:
            if B.involved_in(Q,last_require=2):
              is_good = False
              insertion_locations[Q[-1]] = 0
              # break
          if is_good:

            add_this_time.append(Q)
        for Q in add_this_time:
          # print Q,'is good'
          # print '\tchanging IL from ',Q.insertion_locations,'to',(insertion_locations[:Q[-1]+1]+  insertion_locations[Q[-1]:])
          Q.insertion_locations = insertion_locations[:Q[-1]+1] + insertion_locations[Q[-1]:]
          self[n].add(Q)

  def extend_to_length(self, l):
    for i in range(self.length+1, l+1):
      self.append(PermSet())
    if (l <= self.length):
      return
    old = self.length
    self.length = l
    for n in range(old+1,l+1):
      for P in self[n-1]:
        insertion_locations = P.insertion_locations
        add_this_time = []
        for Q in P.right_extensions():
          is_good = True
          for B in self.basis:
            if B.involved_in(Q,last_require=2):
              is_good = False
              insertion_locations[Q[-1]] = 0
              # break
          if is_good:

            add_this_time.append(Q)
        for Q in add_this_time:
          # print Q,'is good'
          # print '\tchanging IL from ',Q.insertion_locations,'to',(insertion_locations[:Q[-1]+1]+  insertion_locations[Q[-1]:])
          Q.insertion_locations = insertion_locations[:Q[-1]+1] + insertion_locations[Q[-1]:]
          self[n].add(Q)

  def right_juxtaposition(self, C, generate_perms=True):
    A = PermSet()
    max_length = max([len(P) for P in self.basis]) + max([len(P) for P in C.basis])
    for n in range(2, max_length+1):
      for i in range(0, factorial(n)):
        P = Permutation(i,n)
        for Q in self.basis:
          for R in C.basis:
            if len(Q) + len(R) == n:
              if (Q == Permutation(P[0:len(Q)]) and R == Permutation(P[len(Q):n])):
                A.add(P)
            elif len(Q) + len(R) - 1 == n:
              if (Q == Permutation(P[0:len(Q)]) and Permutation(R) == Permutation(P[len(Q)-1:n])):
                A.add(P)
    return AvClass(list(A.minimal_elements()), length=(8 if generate_perms else 0))

  def above_juxtaposition(self, C, generate_perms=True):
    inverse_class = AvClass([P.inverse() for P in C.basis], 0)
    horizontal_juxtaposition = self.right_juxtaposition(inverse_class, generate_perms=False)
    return AvClass([B.inverse() for B in horizontal_juxtaposition.basis], length=(8 if generate_perms else 0))

  def contains(self, C):
    for P in self.basis:
      good = False
      for Q in C.basis:
        if P.involved_in(Q):
          good = True
          break
      if not good:
        return False
    return True



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(name='permpy',
      version='0.0.2',
      description='Permutation (patterns) workshop in Python',
      author='Cheyne Homberger and Jay Pantone',
      author_email='cheyne.homberger@gmail.com',
      url='https://github.com/cheyneh/permpy',
      keywords = ['permutations', 'patterns', 'research',
          'enumeration'],
      classifiers=[],
      packages=['permpy', 
                'permpy.InsertionEncoding', 
                'permpy.RestrictedContainer'],
     )




class Vector(tuple):
  def __new__(cls, v):
    for i in range(0, len(v)):
      v[i] = int(v[i])
      assert v[i] >= 0, 'vector entries must be nonnegative integers'
    return tuple.__new__(cls, v)

  def meet(self, V):
  	assert len(V) == len(self), 'vectors must be same length to meet'
  	new_vector = []
  	for i in range(0, len(V)):
  	  new_vector.append(max(V[i], self[i]))
  	return Vector(new_vector)

  def contained_in(self, V):
    assert len(V) == len(self), 'vectors must be same length to check containment'
    for i in range(0, len(V)):
      if self[i] > V[i]:
        return False
    return True

  def norm(self):
    return sum(self)


from permpy.permutation import Permutation

def fixed_points(perm):
    """Returns the number of fixed points of the permutation.

    >>> Permutation(521436).fixed_points()
    3
    """
    sum = 0
    for i in range(perm.__len__()):
        if perm(i) == i:
            sum += 1
    return sum


def skew_decomposable(perm):
    """Determines whether the permutation is expressible as the skew sum of
    two permutations.

    >>> p = Permutation.random(8).direct_sum(Permutation.random(12))
    >>> p.skew_decomposable()
    False
    >>> p.complement().skew_decomposable()
    True
    """

    p = list(perm)
    n = perm.__len__()
    for i in range(1,n):
        if set(range(n-i,n)) == set(p[0:i]):
            return True
    return False

def sum_decomposable(perm):
    """Determines whether the permutation is expressible as the direct sum of
    two permutations.

    >>> p = Permutation.random(4).direct_sum(Permutation.random(15))
    >>> p.sum_decomposable()
    True
    >>> p.reverse().sum_decomposable()
    False
    """

    p = list(perm)
    n = perm.__len__()
    for i in range(1,n):
        if set(range(0,i)) == set(p[0:i]):
            return True
    return False

def num_cycles(perm):
    """Returns the number of cycles in the permutation.

    >>> Permutation(53814276).num_cycles()
    3
    """

    return len(perm.cycle_decomp())



def descent_set(perm):
    """Returns descent set of the permutation

    >>> Permutation(42561873).descent_set()
    [1, 4, 6, 7]
    """

    p = list(perm)
    n = perm.__len__()
    descents = []
    for i in range(1,n):
        if p[i-1] > p[i]:
            descents.append(i)
    return descents

def num_descents(perm):
    """Returns the number of descents of the permutation

    >>> Permutation(42561873).num_descents()
    4
    """
    return len(perm.descent_set())

def ascent_set(perm):
    """Returns the ascent set of the permutation

    >>> Permutation(42561873).ascent_set()
    [2, 3, 5]
    """
    descents = perm.descent_set()
    return [i for i in range(1, len(perm)) if i not in descents]

def num_ascents(perm):
    """Returns the number of ascents of the permutation

    >>> Permutation(42561873).num_ascents()
    3
    """
    return len(perm.ascent_set())


def peak_list(perm):
    """Returns the list of peaks of the permutation.

    >>> Permutation(2341765).peak_list()
    [2, 4]
    """

    def check(i):
        return perm[i-1] < perm[i] > perm[i+1]
    return [i for i in range(1, len(perm)-1) if check(i)]


def num_peaks(perm):
    """Returns the number of peaks of the permutation

    >>> Permutation(2341765).num_peaks()
    2
    """

    return len(perm.peak_list())

def valley_list(perm):
    """Returns the list of valleys of the permutation.

    >>> Permutation(3241756).valley_list()
    [1, 3, 5]
    """

    return perm.complement().peak_list()


def num_valleys(perm):
    """Returns the number of peaks of the permutation

    >>> Permutation(3241756).num_valleys()
    3
    """

    return len(perm.valley_list())

def bend_list(perm):
    """Returns the list of indices at which the permutation changes
    direction. That is, the number of non-monotone consecutive triples of
    the permutation. A permutation p can be expressed as the concatenation
    of len(p.bend_list()) + 1 monotone segments."""
    raise NotImplementedError

    # this isn't quite correct....
    return len([i for i in range(1, len(perm)-1) if (perm[i-1] > perm[i] and perm[i+1] > perm[i]) or (perm[i-1] < perm[i] and perm[i+1] < perm[i])])


def trivial(perm):
    """The trivial permutation statistic, for convenience

    >>> Permutation.random(10).trivial()
    0
    """
    return 0

def order(perm):
    L = map(len, perm.cycle_decomp())
    return reduce(lambda x,y: x*y // fractions.gcd(x,y), L)

def ltrmin(perm):
    """Returns the positions of the left-to-right minima.

    >>> Permutation(35412).ltrmin()
    [0, 3]
    """

    n = perm.__len__()
    L = []
    minval = len(perm) + 1
    for idx, val in enumerate(perm):
        if val < minval:
            L.append(idx)
            minval = val
    return L

def rtlmin(perm):
    """Returns the positions of the left-to-right minima.

    >>> Permutation(315264).rtlmin()
    [5, 3, 1]
    """
    rev_perm = perm.reverse()
    return [len(perm) - val - 1 for val in rev_perm.ltrmin()]

def ltrmax(perm):
    return [len(perm)-i-1 for i in Permutation(perm[::-1]).rtlmax()][::-1]

def rtlmax(perm):
    return [len(perm)-i-1 for i in perm.complement().reverse().ltrmin()][::-1]

def num_ltrmin(perm):
    return len(perm.ltrmin())

def inversions(perm):
    """Returns the number of inversions of the permutation, i.e., the
    number of pairs i,j such that i < j and perm(i) > perm(j).

    >>> Permutation(4132).inversions()
    4
    >>> Permutation.monotone_decreasing(6).inversions() == 5*6 / 2
    True
    >>> Permutation.monotone_increasing(7).inversions()
    0
    """

    p = list(perm)
    n = perm.__len__()
    inv = 0
    for i in range(n):
        for j in range(i+1,n):
            if p[i]>p[j]:
                inv+=1
    return inv

def noninversions(perm):
    p = list(perm)
    n = perm.__len__()
    inv = 0
    for i in range(n):
        for j in range(i+1,n):
            if p[i]<p[j]:
                inv+=1
    return inv

def bonds(perm):
    numbonds = 0
    p = list(perm)
    for i in range(1,len(p)):
        if p[i] - p[i-1] == 1 or p[i] - p[i-1] == -1:
            numbonds+=1
    return numbonds

def majorindex(perm):
    sum = 0
    p = list(perm)
    n = perm.__len__()
    for i in range(0,n-1):
        if p[i] > p[i+1]:
            sum += i + 1
    return sum

def fixedptsplusbonds(perm):
    return perm.fixed_points() + perm.bonds()

def longestrunA(perm):
    p = list(perm)
    n = perm.__len__()
    maxi = 0
    length = 1
    for i in range(1,n):
        if p[i-1] < p[i]:
            length += 1
            if length > maxi: maxi = length
        else:
            length = 1
    return max(maxi,length)

def longestrunD(perm):
    return perm.complement().longestrunA()

def longestrun(perm):
    return max(perm.longestrunA(), perm.longestrunD())

def christiecycles(perm):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product. used for transpositions
    p = list(perm)
    n = perm.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = range(1,n+1) + [0]
    blackperm = [0 for i in range(n+1)]
    for i in range(n+1):
        ind = q.index(i)
        blackperm[i] = q[ind-1]
    newperm = []
    for i in range(n+1):
        k = blackperm[i]
        j = grayperm[k]
        newperm.append(j)
    return Permutation(newperm).numcycles()

def othercycles(perm):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product
    p = list(perm)
    n = perm.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = [n] + range(n)
    blackperm = [0 for i in range(n+1)]
    for i in range(n+1):
        ind = q.index(i)
        blackperm[i] = q[ind-1]
    newperm = []
    for i in range(n+1):
        k = blackperm[i]
        j = grayperm[k]
        newperm.append(j)
    return Permutation(newperm).numcycles()

def sumcycles(perm):
    return perm.othercycles() + perm.christiecycles()

def maxcycles(perm):
    return max(perm.othercycles() - 1,perm.christiecycles())

def is_involution(perm):
    """Checks if the permutation is an involution, i.e., is equal to it's
    own inverse. """

    return perm == perm.inverse()

def is_identity(perm):
    """Checks if the permutation is the identity.

    >>> p = Permutation.random(10)
    >>> (p * p.inverse()).is_identity()
    True
    """

    return perm == Permutation.identity(len(perm))

def threepats(perm):
    p = list(perm)
    n = perm.__len__()
    patnums = {'123' : 0, '132' : 0, '213' : 0,
                         '231' : 0, '312' : 0, '321' : 0}
    for i in range(n-2):
        for j in range(i+1,n-1):
            for k in range(j+1,n):
                patnums[''.join(map(lambda x:
                                                        str(x+1),Permutation([p[i], p[j], p[k]])))] += 1
    return patnums

def fourpats(perm):
    p = list(perm)
    n = perm.__len__()
    patnums = {'1234' : 0, '1243' : 0, '1324' : 0,
                         '1342' : 0, '1423' : 0, '1432' : 0,
                         '2134' : 0, '2143' : 0, '2314' : 0,
                         '2341' : 0, '2413' : 0, '2431' : 0,
                         '3124' : 0, '3142' : 0, '3214' : 0,
                         '3241' : 0, '3412' : 0, '3421' : 0,
                         '4123' : 0, '4132' : 0, '4213' : 0,
                         '4231' : 0, '4312' : 0, '4321' : 0 }

    for i in range(n-3):
        for j in range(i+1,n-2):
            for k in range(j+1,n-1):
                for m in range(k+1,n):
                    patnums[''.join(map(lambda x:
                                        str(x+1),list(Permutation([p[i], p[j], p[k], p[m]]))))] += 1
    return patnums

def num_consecutive_3214(perm):
    number = 0
    n = len(perm)
    for i in range(n-3):
        if perm[i+2] < perm[i+1] < perm[i] < perm[i+3]:
            number += 1
    return number
